# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_docs.ipynb.

# %% ../nbs/00_docs.ipynb 3
from __future__ import annotations
from .import_essentials import *
import nbdev
from fastcore.basics import AttrDict
from fastcore.utils import *

from nbdev.showdoc import *
from nbdev.doclinks import *
from inspect import isclass
from nbdev.showdoc import (
    _ext_link, 
    _wrap_sig, 
    _fmt_anno, 
    _f_name, 
    DocmentTbl, 
    _maybe_nm, 
    _show_param
)
from nbdev.config import get_config

# %% auto 0
__all__ = ['ListDocment', 'CustomizedMarkdownRenderer']

# %% ../nbs/00_docs.ipynb 4
def _docment_parser(parser: BaseParser):
    p = parser.schema()['properties']
    if hasattr(parser, '__annotations__'):
        anno = parser.__annotations__
    else:
        anno = { k: inspect._empty for k in p.keys() }
    d = { 
        k: {
            'anno': anno[k],
            'default': v['default'] if 'default' in v else inspect._empty,
            'docment': v['description'] if 'description' in v else inspect._empty,
        } for k, v in p.items()
    }
    

    d = AttrDict(d)
    return d


# %% ../nbs/00_docs.ipynb 5
class ParserMarkdownRenderer(BasicMarkdownRenderer):
    def __init__(self, sym, name: str | None = None, title_level: int = 3):
        super().__init__(sym, name, title_level)
        self.dm.dm = _docment_parser(sym)

# %% ../nbs/00_docs.ipynb 6
def _italic(s: str): return f'<em>{s}</em>' if s.strip() else s

def _bold(s: str): return f'<b>{s}</b>' if s.strip() else s

# %% ../nbs/00_docs.ipynb 7
def _show_param(param):
    "Like `Parameter.__str__` except removes: quotes in annos, spaces, ids in reprs"
    kind,res,anno,default = param.kind,param._name,param._annotation,param._default
    kind = '*' if kind==inspect._VAR_POSITIONAL else '**' if kind==inspect._VAR_KEYWORD else ''
    res = kind+res
    # if anno is not inspect._empty: res += f':{_f_name(anno) or _fmt_anno(anno)}'
    if default is not inspect._empty: res += f'={_f_name(default) or repr(default)}'
    return res


def _fmt_sig(sig):
    if sig is None: return ''
    p = {k:v for k,v in sig.parameters.items()}
    _params = [_show_param(p[k]) for k in p.keys() if k != 'self']
    return "(" + ', '.join(_params)  + ")"


# %% ../nbs/00_docs.ipynb 8
def _inner_list2mdlist(l: list):
    param_name, param_anno, param_default, param_doc = l
    # annotation
    if param_anno == inspect._empty: param_anno = None
    else: param_anno = f"`{param_anno}`"
    # default value
    if param_default == inspect._empty: param_default = None
    else: param_default = _italic(f"default={param_default}")

    mdoc = ""
    if param_anno and param_default:
        mdoc += f"* {_bold(param_name)} ({param_anno}, {param_default})"
    elif param_anno:
        mdoc += f"* {_bold(param_name)} ({param_anno})"
    elif param_default:
        mdoc += f"* {_bold(param_name)} ({param_default})"
    else:
        mdoc += f"* {_bold(param_name)}"
    
    if not (param_doc == inspect._empty): 
        mdoc += f" -- {param_doc}"
    return mdoc

def _params_mdlist(tbl: DocmentTbl):
    param_list = [
        L([k, v['anno'], v['default'], v['docment']])
        for k, v in tbl.dm.items() if k != 'return'
    ]
    # param_list = tbl._row_list
    return L(param_list).map(_inner_list2mdlist)

def _return_mdlist(tbl: DocmentTbl):
    return_list = [tbl.dm['return'][k] for k in ['anno', 'default', 'docment']]
    param_anno, param_default, param_doc = return_list
    mdoc = ""
    if not param_anno == inspect._empty: 
        mdoc += f"(`{param_anno}`)"
    if param_doc != inspect._empty:
        mdoc += f" -- {param_doc}"
    return mdoc

def _show_params_return(tbl: DocmentTbl):
    if not tbl.has_docment: return ''
    doc = "" 
    doc = "::: {#docs}\n\n"
    doc += '**Parameters:**' + '\n\n\n\n'
    doc += _params_mdlist(tbl)
    doc += "\n\n:::\n\n"
    if tbl.has_return:
        doc += "::: {#docs}\n\n"
        doc += '\n\n**Returns:**\n'
        doc += f"&ensp;&ensp;&ensp;&ensp;{_return_mdlist(tbl)}"
        doc += "\n\n:::"
    
    return '\n'.join(doc)

# %% ../nbs/00_docs.ipynb 9
class ListDocment:
    def __init__(self, tbl: DocmentTbl):
        self.tbl = tbl
    
    def _repre_mardown(self):
        return _show_params_return(self.tbl)

    __str__ = _repre_mardown

# %% ../nbs/00_docs.ipynb 10
def _repr_markdown(
    renderer: ShowDocRenderer,
    use_module_dir: bool,
    show_title: bool,
    is_class: bool,
    title_level: int = 3,
):
    doc = ""
    src = NbdevLookup().code(renderer.fn)
    _look_up = NbdevLookup()[renderer.fn]
    module_dir = _look_up[1].replace('.py', '').replace('/', '.') + '.' if _look_up else ""
    link = _ext_link(src, '[source]', 'style="float:right; font-size:smaller"') + '\n\n' if src else ""
    
    name = f"{module_dir}{_bold(renderer.nm)}" if use_module_dir else _bold(renderer.nm)

    # title
    if show_title:
        h = f'h{title_level}'
        doc += f"""<{h} class="doc-title" id="{renderer.nm}">{name}</{h}>"""
    # signature
    doc += link
    if is_class: doc += '::: {.doc-sig}\n\n class '
    else: doc += '::: {.doc-sig}\n\n '
    doc += f"{name} {_italic(_fmt_sig(renderer.sig))}\n\n:::"
    # docs
    if renderer.docs: doc += f"\n\n{renderer.docs}"
    # params and return
    if renderer.dm.has_docment:
        doc += f"\n\n{ListDocment(renderer.dm)}"
    return doc


# %% ../nbs/00_docs.ipynb 11
class CustomizedMarkdownRenderer(ShowDocRenderer):
    """Displaying documents of functions, classes, `haiku.module`, and `BaseParser`."""
    
    def __init__(self, sym, name:str|None=None, title_level:int=3):
        super().__init__(sym, name, title_level)
        self.isclass = inspect.isclass(sym)
        self.hook_methods(sym)
        self._check_sym(sym)

    def hook_methods(self, sym):
        self.methods = []
        if self.isclass and hasattr(sym, '__ALL__'):
            all_methods_syms_names = [
                (getattr(sym, x), x) for x in sym.__ALL__
            ]
            self.methods = [ShowDocRenderer(sym, name=str(x)) for sym, x in all_methods_syms_names]

    def _check_sym(self, sym):
       
        if self.isclass:
            # extract annotations for pydantic models
            if issubclass(sym, BaseParser):
                self.dm.dm = _docment_parser(sym)
            # extract annotations for hk.Module
            # if issubclass(sym, hk.Module):
            #     _sym = sym.__init__
            #     try: self.sig = signature_ex(_sym, eval_str=True)
            #     except (ValueError,TypeError): self.sig = None
            #     self.dm = DocmentTbl(_sym)

    def _repr_markdown_(self):
        doc = _repr_markdown(
            self,
            use_module_dir=True,
            show_title=True,
            is_class=self.isclass,
            title_level=self.title_level + 1,
        )
        if self.methods:
            doc += '\n\n::: {.doc-methods} \n\n**Methods** \n\n' 
            doc += '\n\n'.join([
                _repr_markdown(
                    x, use_module_dir=False,
                    show_title=False, is_class=False,
                ) 
                for x in self.methods])
            
            doc += '\n\n:::\n\n'
            
        return doc



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/methods/09_l2c.ipynb.

# %% ../../nbs/methods/09_l2c.ipynb 3
from __future__ import annotations
from ..import_essentials import *
from .base import ParametricCFModule
from ..base import BaseConfig
from ..utils import *
from ..data_utils import Feature, FeaturesList
from ..ml_model import MLP, MLPBlock
from ..data_module import DataModule
from keras.random import SeedGenerator

# %% auto 0
__all__ = ['gumbel_softmax', 'sample_categorical', 'sample_bernouli', 'split_fn', 'L2CModel', 'qcut', 'qcut_inverse',
           'cut_quantiles', 'discretize_xs', 'Discretizer', 'L2CConfig', 'L2C']

# %% ../../nbs/methods/09_l2c.ipynb 7
def gumbel_softmax(
    key: jrand.PRNGKey, # Random key
    logits: Array, # Logits for each class. Shape (batch_size, num_classes)
    tau: float, # Temperature for the Gumbel softmax
):
    """The Gumbel softmax function."""

    gumbel_noise = jrand.gumbel(key, shape=logits.shape)
    y = logits + gumbel_noise
    return jax.nn.softmax(y / tau, axis=-1)

# %% ../../nbs/methods/09_l2c.ipynb 8
def sample_categorical(
    key: jrand.PRNGKey, # Random key
    logits: Array, # Logits for each class. Shape (batch_size, num_classes)
    tau: float, # Temperature for the Gumbel softmax
    training: bool = True, # Apply gumbel softmax if training
):
    """Sample from a categorical distribution."""

    def sample_cat(key, logits):
        cat = jrand.categorical(key, logits=logits, axis=-1)
        return jax.nn.one_hot(cat, logits.shape[-1])

    return lax.cond(
        training,
        lambda _: gumbel_softmax(key, logits, tau=tau),
        lambda _: sample_cat(key, logits),
        None,
    )

# %% ../../nbs/methods/09_l2c.ipynb 10
def sample_bernouli(
    key: jrand.PRNGKey, # Random key
    prob: Array, # Logits for each class. Shape (batch_size, 1)
    tau: float, # Temperature for the Gumbel softmax
    training: bool = True, # Apply gumbel softmax if training
) -> Array:
    """"Sample from a bernouli distribution."""

    def sample_ber(key, prob):
        return jrand.bernoulli(key, p=prob).astype(prob.dtype)
    
    def gumbel_ber(key, prob, tau):
        key_1, key_2 = jrand.split(key)
        gumbel_1 = jrand.gumbel(key_1, shape=prob.shape)
        gumbel_2 = jrand.gumbel(key_2, shape=prob.shape)
        no_logits = (prob * jnp.exp(gumbel_1)) / tau
        de_logits = no_logits + ((1. - prob) * jnp.exp(gumbel_2)) / tau
        return no_logits / de_logits
    
    return lax.cond(
        training,
        lambda _: gumbel_ber(key, prob, tau),
        lambda _: sample_ber(key, prob),
        None,
    )

# %% ../../nbs/methods/09_l2c.ipynb 11
def split_fn(feature_indices: list[tuple[int, int]]):
    feature_indices = tuple([x[1] for x in feature_indices[:-1]])

    @ft.partial(jit, static_argnums=1)
    def split_xs(xs, feature_indices):
        return jnp.split(xs, list(feature_indices), axis=-1)
    
    @ft.partial(jit, static_argnums=1)
    def split_prob(prob, feature_indices):
        return jnp.split(prob, len(feature_indices) + 1, axis=-1)
    
    return ft.partial(split_xs, feature_indices=feature_indices), ft.partial(split_prob, feature_indices=feature_indices)

# %% ../../nbs/methods/09_l2c.ipynb 13
class L2CModel(keras.Model):

    def __init__(
        self,
        generator_layers: list[int],
        selector_layers: list[int],
        feature_indices: list[tuple[int, int]] = None,
        immutable_mask: Array = None,
        pred_fn: Callable = None,
        alpha: float = 1e-4, # Sparsity regularization
        tau: float = 0.7,
        seed: int = None,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.generator_layers = generator_layers
        self.selector_layers = selector_layers
        self.pred_fn = pred_fn
        self.tau = tau
        self.alpha = alpha
        seed = seed or get_config().global_seed
        self.seed_generator = SeedGenerator(seed)
        self.set_features_info(feature_indices)
        self.set_immutable_mask(immutable_mask)
        
        # split functions
        self.split_xs_fn, self.split_prob_fn = split_fn(self.feature_indices)

    @property
    def start_end_indices(self):
        feature_indices = jnp.array(list(map(lambda x: list(x), self.feature_indices)))
        return feature_indices[:-1, 1]

    def set_features_info(self, feature_indices: list[tuple[int, int]]):
        self.feature_indices = feature_indices
        # self.feature_indices = jnp.array(
        #     list(map(lambda x: list(x), feature_indices)))
        # assert self.feature_indices.shape == (len(feature_indices), 2)
        # TODO: check if the feature indices are valid

    def set_immutable_mask(self, immutable_mask: Array):
        self.immutable_mask = immutable_mask

    def set_pred_fn(self, pred_fn: Callable):
        self.pred_fn = pred_fn

    def build(self, input_shape):
        n_feats = len(self.feature_indices)
        self.generator = MLP(
            sizes=self.generator_layers,
            output_size=input_shape[-1],
            dropout_rate=0.0,
            last_activation="linear",
        )
        self.selector = MLP(
            sizes=self.selector_layers,
            output_size=n_feats,
            dropout_rate=0.0,
            last_activation="sigmoid",
        )

    def compute_l2c_loss(self, inputs, cfs, probs):
        # inputs = self.split_xs_fn(inputs)
        # cfs = self.split_xs_fn(cfs)
        y_target = self.pred_fn(inputs).argmin(axis=-1)
        y_pred = self.pred_fn(cfs)
        validity_loss = keras.losses.sparse_categorical_crossentropy(
            y_target, y_pred
        ).mean()
        sparsity = jnp.linalg.norm(probs, ord=1) * self.alpha
        return validity_loss, sparsity
        
    def forward(self, rng_key, inputs, training=False):
        
        def perturb(x, cf, prob, immutable):
            cf = sample_categorical(
                key_2, cf, tau=tau, training=training
            )
            prob = prob * (1 - immutable)
            return x * (1 - prob) + cf * prob

        key_1, key_2 = jrand.split(rng_key)
        select_probs = self.selector(inputs, training=training)
        probs = sample_bernouli(
            key_1, select_probs, 
            tau=self.tau, training=training
        )
        cfs_logits = self.generator(inputs, training=training)
                
        # xs = jnp.split(inputs, start_end, axis=-1)
        # cfs = jnp.split(cfs_logits, start_end, axis=-1)
        # probs = jnp.split(prob, len(self.feature_indices), axis=-1)
        xs = self.split_xs_fn(inputs)
        cfs = self.split_xs_fn(cfs_logits)
        probs = self.split_prob_fn(probs)
        immutables = self.split_prob_fn(self.immutable_mask)
        tau = self.tau
        
        cfs = jax.tree_util.tree_map(
            perturb, xs, cfs, probs, immutables #self.tau, training
        )
        cfs = jnp.concatenate(cfs, axis=-1)
        probs = jnp.concatenate(probs, axis=-1)
        return cfs, probs
    
    def call(self, inputs, training=False):
        rng_key = self.seed_generator.next()
        cfs, probs = self.forward(rng_key, inputs, training=training)
        # loss = self.compute_l2c_loss(inputs, cfs, probs)

        validity_loss, sparsity = self.compute_l2c_loss(inputs, cfs, probs)
        self.add_loss(validity_loss)
        self.add_loss(sparsity)
        return cfs   

# %% ../../nbs/methods/09_l2c.ipynb 15
def qcut(
    x: Array, # Input array
    q: int, # Number of quantiles
    axis: int = 0, # Axis to quantile
) -> tuple[Array, Array]: # (digitized array, quantiles)
    """Quantile binning."""
    
    # Handle edge cases: empty array or single element
    if x.size <= 1:
        return jnp.zeros_like(x), jnp.array([])
    quantiles = jnp.quantile(x, jnp.linspace(0, 1, q + 1)[1:-1], axis=axis)
    digitized = jnp.digitize(x, quantiles)
    return digitized, quantiles

# %% ../../nbs/methods/09_l2c.ipynb 17
def qcut_inverse(
    digitized: Array, # Digitized One-Hot Encoding Array
    quantiles: Array, # Quantiles
) -> Array:
    """Inverse of qcut."""
    
    result = digitized @ quantiles
    if result.ndim == 1:
        result = result[..., None]
    return result

# %% ../../nbs/methods/09_l2c.ipynb 19
def cut_quantiles(
    quantiles: Array, # Quantiles
    xs: Array, # Input array
):
    quantiles = jnp.concatenate([
        xs.min(axis=0, keepdims=True), 
        quantiles, 
        xs.max(axis=0, keepdims=True)
    ])
    quantiles = (quantiles[1:] + quantiles[:-1]) / 2
    return quantiles

# %% ../../nbs/methods/09_l2c.ipynb 20
def discretize_xs(
    xs: Array, # Input array
    is_categorical_and_indices: list[tuple[bool, tuple[int, int]]], # Features list
    q: int = 4, # Number of quantiles
) -> tuple[list[Array], list[Array], list[Array], list[list[int, int]]]: # (discretized array, indices_and_quantiles_and_mid)
    """Discretize continuous features."""
    
    discretized_xs = []
    mid_quantiles = []
    quantiles_feats = []
    feature_indices = []
    discretized_start, discretized_end = 0, 0

    for is_categorical, (start, end) in is_categorical_and_indices:
        if is_categorical:
            discretized, quantiles, mid = xs[:, start:end], None, jnp.identity(end - start)
            discretized_end += end - start
        else:
            discretized, quantiles = qcut(xs[:, start:end].reshape(-1), q=q)
            mid = cut_quantiles(quantiles, xs[:, start])
            discretized = jax.nn.one_hot(discretized, q)
            discretized_end += discretized.shape[-1]
        
        discretized_xs.append(discretized)
        quantiles_feats.append(quantiles)
        mid_quantiles.append(mid)
        feature_indices.append([discretized_start, discretized_end])
        discretized_start = discretized_end
    # discretized_xs = jnp.concatenate(discretized_xs, axis=-1)
    return discretized_xs, quantiles_feats, mid_quantiles, feature_indices

# %% ../../nbs/methods/09_l2c.ipynb 22
class Discretizer:
    """Discretize continuous features."""
    
    def __init__(
        self, 
        is_cat_and_indices: list[tuple[bool, tuple[int, int]]], # Features list
        q: int = 4 # Number of quantiles
    ):
        self.is_cat_and_indices = is_cat_and_indices
        self.q = q

    @property
    def transform_indices(self):
        return [x[1][1] for x in self.is_cat_and_indices[:-1]]
    
    @property
    def inverse_transform_indices(self):
        return [x[1] for x in self.indices[:-1]]
        
    def fit(self, xs: Array):
        _, self.quantiles, self.mid_quantiles, self.indices = discretize_xs(
            xs, self.is_cat_and_indices, self.q
        )
        self.transform_indices, self.inverse_transform_indices
        return self

    # @ft.partial(jit, static_argnums=0)
    def transform(self, xs: Array):
        def digitize_fn(x, quantile):
            if quantile is None: 
                return x
            else: 
                digitized = jnp.digitize(x.reshape(-1), quantile)
                return jax.nn.one_hot(digitized, self.q)

        # indices = [x[1][1] for x in self.is_cat_and_indices[:-1]]
        # print(indices)
        digitized_xs = jnp.split(xs, self.transform_indices, axis=-1) # [feat_1, ..., feat_n]
        digitized_xs = jax.tree_util.tree_map(
            digitize_fn, digitized_xs, self.quantiles
        )        
        return jnp.concatenate(digitized_xs, axis=-1)

    def fit_transform(self, xs: Array):
        self.fit(xs)
        return self.transform(xs)

    # @ft.partial(jit, static_argnums=0)
    def inverse_transform(self, xs: Array):
        xs = jnp.split(xs, self.inverse_transform_indices, axis=-1)
        xs = jax.tree_util.tree_map(
            lambda x, q: qcut_inverse(x, q), xs, self.mid_quantiles
        )
        return jnp.concatenate(xs, axis=-1)
    
    def inversed_transform_pytree(self, xs: list[Array]):
        xs = jax.tree_util.tree_map(
            lambda x, q: qcut(x, q), xs, self.mid_quantiles)
        return jnp.concatenate(xs, axis=-1)
    
    def get_pred_fn(self, pred_fn: Callable[[Array], Array]):
        def _pred_fn(xs: Array):
            return pred_fn(self.inverse_transform(xs))
            # return pred_fn(self.inversed_transform_pytree(xs))
        return _pred_fn


# %% ../../nbs/methods/09_l2c.ipynb 25
class L2CConfig(BaseConfig):
    generator_layers: list[int] = Field(
        [64, 64, 64], description="Generator MLP layers."
    )
    selector_layers: list[int] = Field(
        [64], description="Selector MLP layers."
    )
    lr: float = Field(1e-3, description="Model learning rate.")
    opt_name: str = Field("adam", description="Optimizer name of training L2C.")
    alpha: float = Field(1e-4, description="Sparsity regularization.")
    tau: float = Field(0.7, description="Temperature for the Gumbel softmax.")
    q: int = Field(4, description="Number of quantiles.")

# %% ../../nbs/methods/09_l2c.ipynb 26
class L2C(ParametricCFModule):
    def __init__(
        self,
        config: Dict | L2CConfig = None,
        l2c_model: L2CModel = None,
        name: str = "l2c",
    ):
        if config is None:
            config = L2CConfig()
        config = validate_configs(config, L2CConfig)
        name = name or "l2c"
        self.l2c_model = l2c_model
        super().__init__(config=config, name=name)

    def train(
        self, 
        data: DataModule, 
        pred_fn: Callable,
        batch_size: int = 128,
        epochs: int = 10,
        **fit_kwargs
    ):
        if not isinstance(data, DataModule):
            raise ValueError(f"Only support `data` to be `DataModule`, "
                             f"got type=`{type(data).__name__}` instead.")
        
        xs_train, ys_train = data['train']
        self.discretizer = Discretizer(
            [(feat.is_categorical, indices) for feat, indices in zip(data.features, data.features.feature_indices)],
            q=self.config.q
        )
        discretized_xs_train = self.discretizer.fit_transform(xs_train)
        pred_fn = self.discretizer.get_pred_fn(pred_fn)
        features_indices = self.discretizer.indices

        self.l2c_model = L2CModel(
            generator_layers=self.config.generator_layers,
            selector_layers=self.config.selector_layers,
            feature_indices=features_indices,
            immutable_mask=jnp.array([feat.is_immutable for feat in data.features], dtype=jnp.int32),
            pred_fn=pred_fn,
            alpha=self.config.alpha,
            tau=self.config.tau,
        )
        self.l2c_model.compile(
            optimizer=keras.optimizers.get({
                'class_name': self.config.opt_name, 
                'config': {'learning_rate': self.config.lr}
            }),
            loss=None,
        )
        self.l2c_model.fit(
            discretized_xs_train, ys_train,
            epochs=epochs,
            batch_size=batch_size,
            **fit_kwargs
        )
        self._is_trained = True
        return self
    
    @auto_reshaping('x')
    def generate_cf(
        self,
        x: Array,
        pred_fn: Callable = None,
        y_target: Array = None,
        rng_key: jrand.PRNGKey = None,
        **kwargs
    ) -> Array:
        
        @jax.jit
        def generate_cf(x: Array):
            discretized_x = self.discretizer.transform(x)
            cfs, probs = self.l2c_model.forward(rng_key, discretized_x, training=False)
            return self.discretizer.inverse_transform(cfs)
        return generate_cf(x)



"""Manage the model and optimizer checkpoints"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/legacy/04b_ckpt_manager.ipynb.

# %% auto 0
__all__ = ['save_checkpoint', 'load_checkpoint', 'CheckpointManager']

# %% ../../nbs/legacy/04b_ckpt_manager.ipynb 3
from .import_essentials import *
from collections import OrderedDict

# %% ../../nbs/legacy/04b_ckpt_manager.ipynb 4
# https://github.com/deepmind/dm-haiku/issues/18#issuecomment-981814403
def save_checkpoint(state, ckpt_dir: Path):
    with open(os.path.join(ckpt_dir, "params.npy"), "wb") as f:
        for x in jax.tree_util.tree_leaves(state):
            np.save(f, x, allow_pickle=False)

    tree_struct = jax.tree_util.tree_map(lambda t: 0, state)
    with open(os.path.join(ckpt_dir, "tree.pkl"), "wb") as f:
        pickle.dump(tree_struct, f)


def load_checkpoint(ckpt_dir: Path):
    with open(os.path.join(ckpt_dir, "tree.pkl"), "rb") as f:
        tree_struct = pickle.load(f)

    leaves, treedef = jax.tree_util.tree_flatten(tree_struct)
    with open(os.path.join(ckpt_dir, "params.npy"), "rb") as f:
        flat_state = [np.load(f) for _ in leaves]

    return jax.tree_util.tree_unflatten(treedef, flat_state)


# %% ../../nbs/legacy/04b_ckpt_manager.ipynb 5
class CheckpointManager:
    def __init__(
        self,
        log_dir: Union[Path, str],
        monitor_metrics: Optional[str],
        max_n_checkpoints: int = 3,
    ):
        self.log_dir = Path(log_dir)
        self.monitor_metrics = monitor_metrics
        self.max_n_checkpoints = max_n_checkpoints
        self.checkpoints = OrderedDict()
        self.n_checkpoints = 0
        if self.monitor_metrics is None:
            warnings.warn(
                "`monitor_metrics` is not specified in `CheckpointManager`. No checkpoints will be stored."
            )

    # update checkpoints based on monitor_metrics
    def update_checkpoints(
        self,
        params: hk.Params,
        opt_state: optax.OptState,
        epoch_logs: Dict[str, float],
        epochs: int,
        steps: Optional[int] = None,
    ):
        if self.monitor_metrics is None:
            return
        if self.monitor_metrics not in epoch_logs:
            raise ValueError(
                "The monitor_metrics ({}) is not appropriately configured.".format(
                    self.monitor_metrics
                )
            )
        metric = float(epoch_logs[self.monitor_metrics])
        if steps:
            ckpt_name = f"epoch={epochs}_step={steps}"
        else:
            ckpt_name = f"epoch={epochs}"

        if self.n_checkpoints < self.max_n_checkpoints:
            self.checkpoints[metric] = ckpt_name
            self.save_net_opt(params, opt_state, ckpt_name)
            self.n_checkpoints += 1
        else:
            old_metric, old_ckpt_name = self.checkpoints.popitem(last=True)
            if metric < old_metric:
                self.checkpoints[metric] = ckpt_name
                self.save_net_opt(params, opt_state, ckpt_name)
                self.delete_net_opt(old_ckpt_name)
            else:
                self.checkpoints[old_metric] = old_ckpt_name

        self.checkpoints = OrderedDict(
            sorted(self.checkpoints.items(), key=lambda x: x[0])
        )

    def save_net_opt(self, params, opt_state, ckpt_name: str):
        ckpt_dir = self.log_dir / f"{ckpt_name}"
        ckpt_dir.mkdir(parents=True, exist_ok=True)
        model_ckpt_dir = ckpt_dir / "model"
        opt_ckpt_dir = ckpt_dir / "opt"
        # create dirs for storing states of model and optimizer
        model_ckpt_dir.mkdir(parents=True, exist_ok=True)
        opt_ckpt_dir.mkdir(parents=True, exist_ok=True)
        # save model and optimizer states
        save_checkpoint(params, model_ckpt_dir)
        save_checkpoint(opt_state, opt_ckpt_dir)

    def delete_net_opt(self, ckpt_name: str):
        ckpt_dir = self.log_dir / f"{ckpt_name}"
        shutil.rmtree(ckpt_dir)



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'master',
                'doc_baseurl': '/jax-relax',
                'doc_host': 'https://birkhoffg.github.io',
                'git_url': 'https://github.com/birkhoffg/jax-relax',
                'lib_path': 'relax'},
  'syms': { 'relax.base': { 'relax.base.BaseConfig': ('base.html#baseconfig', 'relax/base.py'),
                            'relax.base.BaseConfig.load_from_json': ('base.html#baseconfig.load_from_json', 'relax/base.py'),
                            'relax.base.BaseConfig.save': ('base.html#baseconfig.save', 'relax/base.py'),
                            'relax.base.BaseModule': ('base.html#basemodule', 'relax/base.py'),
                            'relax.base.BaseModule.__init__': ('base.html#basemodule.__init__', 'relax/base.py'),
                            'relax.base.BaseModule.load_from_path': ('base.html#basemodule.load_from_path', 'relax/base.py'),
                            'relax.base.BaseModule.name': ('base.html#basemodule.name', 'relax/base.py'),
                            'relax.base.BaseModule.save': ('base.html#basemodule.save', 'relax/base.py'),
                            'relax.base.PredFnMixedin': ('base.html#predfnmixedin', 'relax/base.py'),
                            'relax.base.PredFnMixedin.pred_fn': ('base.html#predfnmixedin.pred_fn', 'relax/base.py'),
                            'relax.base.TrainableMixedin': ('base.html#trainablemixedin', 'relax/base.py'),
                            'relax.base.TrainableMixedin.is_trained': ('base.html#trainablemixedin.is_trained', 'relax/base.py'),
                            'relax.base.TrainableMixedin.train': ('base.html#trainablemixedin.train', 'relax/base.py')},
            'relax.data_module': { 'relax.data_module.BaseDataModule': ('data.html#basedatamodule', 'relax/data_module.py'),
                                   'relax.data_module.BaseDataModule._prepare': ( 'data.html#basedatamodule._prepare',
                                                                                  'relax/data_module.py'),
                                   'relax.data_module.BaseDataModule.apply_constraints': ( 'data.html#basedatamodule.apply_constraints',
                                                                                           'relax/data_module.py'),
                                   'relax.data_module.BaseDataModule.compute_reg_loss': ( 'data.html#basedatamodule.compute_reg_loss',
                                                                                          'relax/data_module.py'),
                                   'relax.data_module.DataModule': ('data.html#datamodule', 'relax/data_module.py'),
                                   'relax.data_module.DataModule.__getitem__': ('data.html#datamodule.__getitem__', 'relax/data_module.py'),
                                   'relax.data_module.DataModule.__init__': ('data.html#datamodule.__init__', 'relax/data_module.py'),
                                   'relax.data_module.DataModule._get_data': ('data.html#datamodule._get_data', 'relax/data_module.py'),
                                   'relax.data_module.DataModule._prepare': ('data.html#datamodule._prepare', 'relax/data_module.py'),
                                   'relax.data_module.DataModule.apply_constraints': ( 'data.html#datamodule.apply_constraints',
                                                                                       'relax/data_module.py'),
                                   'relax.data_module.DataModule.compute_reg_loss': ( 'data.html#datamodule.compute_reg_loss',
                                                                                      'relax/data_module.py'),
                                   'relax.data_module.DataModule.from_config': ('data.html#datamodule.from_config', 'relax/data_module.py'),
                                   'relax.data_module.DataModule.from_features': ( 'data.html#datamodule.from_features',
                                                                                   'relax/data_module.py'),
                                   'relax.data_module.DataModule.from_numpy': ('data.html#datamodule.from_numpy', 'relax/data_module.py'),
                                   'relax.data_module.DataModule.from_path': ('data.html#datamodule.from_path', 'relax/data_module.py'),
                                   'relax.data_module.DataModule.inverse_transform': ( 'data.html#datamodule.inverse_transform',
                                                                                       'relax/data_module.py'),
                                   'relax.data_module.DataModule.load_from_path': ( 'data.html#datamodule.load_from_path',
                                                                                    'relax/data_module.py'),
                                   'relax.data_module.DataModule.sample': ('data.html#datamodule.sample', 'relax/data_module.py'),
                                   'relax.data_module.DataModule.save': ('data.html#datamodule.save', 'relax/data_module.py'),
                                   'relax.data_module.DataModule.set_transformations': ( 'data.html#datamodule.set_transformations',
                                                                                         'relax/data_module.py'),
                                   'relax.data_module.DataModule.transform': ('data.html#datamodule.transform', 'relax/data_module.py'),
                                   'relax.data_module.DataModuleConfig': ('data.html#datamoduleconfig', 'relax/data_module.py'),
                                   'relax.data_module.DataModuleConfig.shuffle': ( 'data.html#datamoduleconfig.shuffle',
                                                                                   'relax/data_module.py'),
                                   'relax.data_module.DataModuleInfoMixin': ('data.html#datamoduleinfomixin', 'relax/data_module.py'),
                                   'relax.data_module.DataModuleInfoMixin.data': ( 'data.html#datamoduleinfomixin.data',
                                                                                   'relax/data_module.py'),
                                   'relax.data_module.DataModuleInfoMixin.dataset': ( 'data.html#datamoduleinfomixin.dataset',
                                                                                      'relax/data_module.py'),
                                   'relax.data_module.DataModuleInfoMixin.features': ( 'data.html#datamoduleinfomixin.features',
                                                                                       'relax/data_module.py'),
                                   'relax.data_module.DataModuleInfoMixin.label': ( 'data.html#datamoduleinfomixin.label',
                                                                                    'relax/data_module.py'),
                                   'relax.data_module.DataModuleInfoMixin.test_indices': ( 'data.html#datamoduleinfomixin.test_indices',
                                                                                           'relax/data_module.py'),
                                   'relax.data_module.DataModuleInfoMixin.train_indices': ( 'data.html#datamoduleinfomixin.train_indices',
                                                                                            'relax/data_module.py'),
                                   'relax.data_module.DataModuleInfoMixin.xs': ('data.html#datamoduleinfomixin.xs', 'relax/data_module.py'),
                                   'relax.data_module.DataModuleInfoMixin.ys': ('data.html#datamoduleinfomixin.ys', 'relax/data_module.py'),
                                   'relax.data_module._validate_dataname': ('data.html#_validate_dataname', 'relax/data_module.py'),
                                   'relax.data_module.dataframe2features': ('data.html#dataframe2features', 'relax/data_module.py'),
                                   'relax.data_module.dataframe2labels': ('data.html#dataframe2labels', 'relax/data_module.py'),
                                   'relax.data_module.dm_equals': ('data.html#dm_equals', 'relax/data_module.py'),
                                   'relax.data_module.download_data_module_files': ( 'data.html#download_data_module_files',
                                                                                     'relax/data_module.py'),
                                   'relax.data_module.features2config': ('data.html#features2config', 'relax/data_module.py'),
                                   'relax.data_module.features2pandas': ('data.html#features2pandas', 'relax/data_module.py'),
                                   'relax.data_module.load_data': ('data.html#load_data', 'relax/data_module.py'),
                                   'relax.data_module.to_feature': ('data.html#to_feature', 'relax/data_module.py')},
            'relax.data_utils.features': { 'relax.data_utils.features.Feature': ( 'data_utils/features.html#feature',
                                                                                  'relax/data_utils/features.py'),
                                           'relax.data_utils.features.Feature.__get_item__': ( 'data_utils/features.html#feature.__get_item__',
                                                                                               'relax/data_utils/features.py'),
                                           'relax.data_utils.features.Feature.__init__': ( 'data_utils/features.html#feature.__init__',
                                                                                           'relax/data_utils/features.py'),
                                           'relax.data_utils.features.Feature.__repr__': ( 'data_utils/features.html#feature.__repr__',
                                                                                           'relax/data_utils/features.py'),
                                           'relax.data_utils.features.Feature._dispatch_transformation': ( 'data_utils/features.html#feature._dispatch_transformation',
                                                                                                           'relax/data_utils/features.py'),
                                           'relax.data_utils.features.Feature._init_is_categorical': ( 'data_utils/features.html#feature._init_is_categorical',
                                                                                                       'relax/data_utils/features.py'),
                                           'relax.data_utils.features.Feature.apply_constraints': ( 'data_utils/features.html#feature.apply_constraints',
                                                                                                    'relax/data_utils/features.py'),
                                           'relax.data_utils.features.Feature.compute_reg_loss': ( 'data_utils/features.html#feature.compute_reg_loss',
                                                                                                   'relax/data_utils/features.py'),
                                           'relax.data_utils.features.Feature.data': ( 'data_utils/features.html#feature.data',
                                                                                       'relax/data_utils/features.py'),
                                           'relax.data_utils.features.Feature.fit': ( 'data_utils/features.html#feature.fit',
                                                                                      'relax/data_utils/features.py'),
                                           'relax.data_utils.features.Feature.fit_transform': ( 'data_utils/features.html#feature.fit_transform',
                                                                                                'relax/data_utils/features.py'),
                                           'relax.data_utils.features.Feature.from_dict': ( 'data_utils/features.html#feature.from_dict',
                                                                                            'relax/data_utils/features.py'),
                                           'relax.data_utils.features.Feature.inverse_transform': ( 'data_utils/features.html#feature.inverse_transform',
                                                                                                    'relax/data_utils/features.py'),
                                           'relax.data_utils.features.Feature.is_categorical': ( 'data_utils/features.html#feature.is_categorical',
                                                                                                 'relax/data_utils/features.py'),
                                           'relax.data_utils.features.Feature.is_immutable': ( 'data_utils/features.html#feature.is_immutable',
                                                                                               'relax/data_utils/features.py'),
                                           'relax.data_utils.features.Feature.set_transformation': ( 'data_utils/features.html#feature.set_transformation',
                                                                                                     'relax/data_utils/features.py'),
                                           'relax.data_utils.features.Feature.to_dict': ( 'data_utils/features.html#feature.to_dict',
                                                                                          'relax/data_utils/features.py'),
                                           'relax.data_utils.features.Feature.transform': ( 'data_utils/features.html#feature.transform',
                                                                                            'relax/data_utils/features.py'),
                                           'relax.data_utils.features.Feature.transformation': ( 'data_utils/features.html#feature.transformation',
                                                                                                 'relax/data_utils/features.py'),
                                           'relax.data_utils.features.Feature.transformed_data': ( 'data_utils/features.html#feature.transformed_data',
                                                                                                   'relax/data_utils/features.py'),
                                           'relax.data_utils.features.Feature.with_transformed_data': ( 'data_utils/features.html#feature.with_transformed_data',
                                                                                                        'relax/data_utils/features.py'),
                                           'relax.data_utils.features.FeaturesList': ( 'data_utils/features.html#featureslist',
                                                                                       'relax/data_utils/features.py'),
                                           'relax.data_utils.features.FeaturesList.__getitem__': ( 'data_utils/features.html#featureslist.__getitem__',
                                                                                                   'relax/data_utils/features.py'),
                                           'relax.data_utils.features.FeaturesList.__init__': ( 'data_utils/features.html#featureslist.__init__',
                                                                                                'relax/data_utils/features.py'),
                                           'relax.data_utils.features.FeaturesList.__iter__': ( 'data_utils/features.html#featureslist.__iter__',
                                                                                                'relax/data_utils/features.py'),
                                           'relax.data_utils.features.FeaturesList.__len__': ( 'data_utils/features.html#featureslist.__len__',
                                                                                               'relax/data_utils/features.py'),
                                           'relax.data_utils.features.FeaturesList.__next__': ( 'data_utils/features.html#featureslist.__next__',
                                                                                                'relax/data_utils/features.py'),
                                           'relax.data_utils.features.FeaturesList._transform_data': ( 'data_utils/features.html#featureslist._transform_data',
                                                                                                       'relax/data_utils/features.py'),
                                           'relax.data_utils.features.FeaturesList.apply_constraints': ( 'data_utils/features.html#featureslist.apply_constraints',
                                                                                                         'relax/data_utils/features.py'),
                                           'relax.data_utils.features.FeaturesList.compute_reg_loss': ( 'data_utils/features.html#featureslist.compute_reg_loss',
                                                                                                        'relax/data_utils/features.py'),
                                           'relax.data_utils.features.FeaturesList.feature_indices': ( 'data_utils/features.html#featureslist.feature_indices',
                                                                                                       'relax/data_utils/features.py'),
                                           'relax.data_utils.features.FeaturesList.feature_name_indices': ( 'data_utils/features.html#featureslist.feature_name_indices',
                                                                                                            'relax/data_utils/features.py'),
                                           'relax.data_utils.features.FeaturesList.features': ( 'data_utils/features.html#featureslist.features',
                                                                                                'relax/data_utils/features.py'),
                                           'relax.data_utils.features.FeaturesList.features_and_indices': ( 'data_utils/features.html#featureslist.features_and_indices',
                                                                                                            'relax/data_utils/features.py'),
                                           'relax.data_utils.features.FeaturesList.from_dict': ( 'data_utils/features.html#featureslist.from_dict',
                                                                                                 'relax/data_utils/features.py'),
                                           'relax.data_utils.features.FeaturesList.inverse_transform': ( 'data_utils/features.html#featureslist.inverse_transform',
                                                                                                         'relax/data_utils/features.py'),
                                           'relax.data_utils.features.FeaturesList.load_from_path': ( 'data_utils/features.html#featureslist.load_from_path',
                                                                                                      'relax/data_utils/features.py'),
                                           'relax.data_utils.features.FeaturesList.save': ( 'data_utils/features.html#featureslist.save',
                                                                                            'relax/data_utils/features.py'),
                                           'relax.data_utils.features.FeaturesList.set_transformations': ( 'data_utils/features.html#featureslist.set_transformations',
                                                                                                           'relax/data_utils/features.py'),
                                           'relax.data_utils.features.FeaturesList.to_dict': ( 'data_utils/features.html#featureslist.to_dict',
                                                                                               'relax/data_utils/features.py'),
                                           'relax.data_utils.features.FeaturesList.to_pandas': ( 'data_utils/features.html#featureslist.to_pandas',
                                                                                                 'relax/data_utils/features.py'),
                                           'relax.data_utils.features.FeaturesList.transform': ( 'data_utils/features.html#featureslist.transform',
                                                                                                 'relax/data_utils/features.py'),
                                           'relax.data_utils.features.FeaturesList.transformed_data': ( 'data_utils/features.html#featureslist.transformed_data',
                                                                                                        'relax/data_utils/features.py'),
                                           'relax.data_utils.features.FeaturesList.with_transformed_data': ( 'data_utils/features.html#featureslist.with_transformed_data',
                                                                                                             'relax/data_utils/features.py')},
            'relax.data_utils.preprocessing': { 'relax.data_utils.preprocessing.DataPreprocessor': ( 'data_utils/preprocessing.html#datapreprocessor',
                                                                                                     'relax/data_utils/preprocessing.py'),
                                                'relax.data_utils.preprocessing.DataPreprocessor.__init__': ( 'data_utils/preprocessing.html#datapreprocessor.__init__',
                                                                                                              'relax/data_utils/preprocessing.py'),
                                                'relax.data_utils.preprocessing.DataPreprocessor.fit': ( 'data_utils/preprocessing.html#datapreprocessor.fit',
                                                                                                         'relax/data_utils/preprocessing.py'),
                                                'relax.data_utils.preprocessing.DataPreprocessor.fit_transform': ( 'data_utils/preprocessing.html#datapreprocessor.fit_transform',
                                                                                                                   'relax/data_utils/preprocessing.py'),
                                                'relax.data_utils.preprocessing.DataPreprocessor.from_dict': ( 'data_utils/preprocessing.html#datapreprocessor.from_dict',
                                                                                                               'relax/data_utils/preprocessing.py'),
                                                'relax.data_utils.preprocessing.DataPreprocessor.inverse_transform': ( 'data_utils/preprocessing.html#datapreprocessor.inverse_transform',
                                                                                                                       'relax/data_utils/preprocessing.py'),
                                                'relax.data_utils.preprocessing.DataPreprocessor.to_dict': ( 'data_utils/preprocessing.html#datapreprocessor.to_dict',
                                                                                                             'relax/data_utils/preprocessing.py'),
                                                'relax.data_utils.preprocessing.DataPreprocessor.transform': ( 'data_utils/preprocessing.html#datapreprocessor.transform',
                                                                                                               'relax/data_utils/preprocessing.py'),
                                                'relax.data_utils.preprocessing.EncoderPreprocessor': ( 'data_utils/preprocessing.html#encoderpreprocessor',
                                                                                                        'relax/data_utils/preprocessing.py'),
                                                'relax.data_utils.preprocessing.EncoderPreprocessor._fit': ( 'data_utils/preprocessing.html#encoderpreprocessor._fit',
                                                                                                             'relax/data_utils/preprocessing.py'),
                                                'relax.data_utils.preprocessing.EncoderPreprocessor._inverse_transform': ( 'data_utils/preprocessing.html#encoderpreprocessor._inverse_transform',
                                                                                                                           'relax/data_utils/preprocessing.py'),
                                                'relax.data_utils.preprocessing.EncoderPreprocessor._transform': ( 'data_utils/preprocessing.html#encoderpreprocessor._transform',
                                                                                                                   'relax/data_utils/preprocessing.py'),
                                                'relax.data_utils.preprocessing.EncoderPreprocessor.from_dict': ( 'data_utils/preprocessing.html#encoderpreprocessor.from_dict',
                                                                                                                  'relax/data_utils/preprocessing.py'),
                                                'relax.data_utils.preprocessing.EncoderPreprocessor.to_dict': ( 'data_utils/preprocessing.html#encoderpreprocessor.to_dict',
                                                                                                                'relax/data_utils/preprocessing.py'),
                                                'relax.data_utils.preprocessing.MinMaxScaler': ( 'data_utils/preprocessing.html#minmaxscaler',
                                                                                                 'relax/data_utils/preprocessing.py'),
                                                'relax.data_utils.preprocessing.MinMaxScaler.__init__': ( 'data_utils/preprocessing.html#minmaxscaler.__init__',
                                                                                                          'relax/data_utils/preprocessing.py'),
                                                'relax.data_utils.preprocessing.MinMaxScaler.fit': ( 'data_utils/preprocessing.html#minmaxscaler.fit',
                                                                                                     'relax/data_utils/preprocessing.py'),
                                                'relax.data_utils.preprocessing.MinMaxScaler.from_dict': ( 'data_utils/preprocessing.html#minmaxscaler.from_dict',
                                                                                                           'relax/data_utils/preprocessing.py'),
                                                'relax.data_utils.preprocessing.MinMaxScaler.inverse_transform': ( 'data_utils/preprocessing.html#minmaxscaler.inverse_transform',
                                                                                                                   'relax/data_utils/preprocessing.py'),
                                                'relax.data_utils.preprocessing.MinMaxScaler.to_dict': ( 'data_utils/preprocessing.html#minmaxscaler.to_dict',
                                                                                                         'relax/data_utils/preprocessing.py'),
                                                'relax.data_utils.preprocessing.MinMaxScaler.transform': ( 'data_utils/preprocessing.html#minmaxscaler.transform',
                                                                                                           'relax/data_utils/preprocessing.py'),
                                                'relax.data_utils.preprocessing.OneHotEncoder': ( 'data_utils/preprocessing.html#onehotencoder',
                                                                                                  'relax/data_utils/preprocessing.py'),
                                                'relax.data_utils.preprocessing.OneHotEncoder.fit': ( 'data_utils/preprocessing.html#onehotencoder.fit',
                                                                                                      'relax/data_utils/preprocessing.py'),
                                                'relax.data_utils.preprocessing.OneHotEncoder.inverse_transform': ( 'data_utils/preprocessing.html#onehotencoder.inverse_transform',
                                                                                                                    'relax/data_utils/preprocessing.py'),
                                                'relax.data_utils.preprocessing.OneHotEncoder.transform': ( 'data_utils/preprocessing.html#onehotencoder.transform',
                                                                                                            'relax/data_utils/preprocessing.py'),
                                                'relax.data_utils.preprocessing.OrdinalPreprocessor': ( 'data_utils/preprocessing.html#ordinalpreprocessor',
                                                                                                        'relax/data_utils/preprocessing.py'),
                                                'relax.data_utils.preprocessing.OrdinalPreprocessor.fit': ( 'data_utils/preprocessing.html#ordinalpreprocessor.fit',
                                                                                                            'relax/data_utils/preprocessing.py'),
                                                'relax.data_utils.preprocessing.OrdinalPreprocessor.inverse_transform': ( 'data_utils/preprocessing.html#ordinalpreprocessor.inverse_transform',
                                                                                                                          'relax/data_utils/preprocessing.py'),
                                                'relax.data_utils.preprocessing.OrdinalPreprocessor.transform': ( 'data_utils/preprocessing.html#ordinalpreprocessor.transform',
                                                                                                                  'relax/data_utils/preprocessing.py'),
                                                'relax.data_utils.preprocessing._check_xs': ( 'data_utils/preprocessing.html#_check_xs',
                                                                                              'relax/data_utils/preprocessing.py'),
                                                'relax.data_utils.preprocessing._unique': ( 'data_utils/preprocessing.html#_unique',
                                                                                            'relax/data_utils/preprocessing.py')},
            'relax.data_utils.transforms': { 'relax.data_utils.transforms.BaseTransformation': ( 'data_utils/transform.html#basetransformation',
                                                                                                 'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms.BaseTransformation.__init__': ( 'data_utils/transform.html#basetransformation.__init__',
                                                                                                          'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms.BaseTransformation.apply_constraints': ( 'data_utils/transform.html#basetransformation.apply_constraints',
                                                                                                                   'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms.BaseTransformation.compute_reg_loss': ( 'data_utils/transform.html#basetransformation.compute_reg_loss',
                                                                                                                  'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms.BaseTransformation.fit': ( 'data_utils/transform.html#basetransformation.fit',
                                                                                                     'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms.BaseTransformation.fit_transform': ( 'data_utils/transform.html#basetransformation.fit_transform',
                                                                                                               'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms.BaseTransformation.from_dict': ( 'data_utils/transform.html#basetransformation.from_dict',
                                                                                                           'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms.BaseTransformation.inverse_transform': ( 'data_utils/transform.html#basetransformation.inverse_transform',
                                                                                                                   'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms.BaseTransformation.is_categorical': ( 'data_utils/transform.html#basetransformation.is_categorical',
                                                                                                                'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms.BaseTransformation.to_dict': ( 'data_utils/transform.html#basetransformation.to_dict',
                                                                                                         'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms.BaseTransformation.transform': ( 'data_utils/transform.html#basetransformation.transform',
                                                                                                           'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms.GumbelSoftmaxTransformation': ( 'data_utils/transform.html#gumbelsoftmaxtransformation',
                                                                                                          'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms.GumbelSoftmaxTransformation.__init__': ( 'data_utils/transform.html#gumbelsoftmaxtransformation.__init__',
                                                                                                                   'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms.GumbelSoftmaxTransformation.apply_constraints': ( 'data_utils/transform.html#gumbelsoftmaxtransformation.apply_constraints',
                                                                                                                            'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms.GumbelSoftmaxTransformation.soft_constraints': ( 'data_utils/transform.html#gumbelsoftmaxtransformation.soft_constraints',
                                                                                                                           'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms.GumbelSoftmaxTransformation.to_dict': ( 'data_utils/transform.html#gumbelsoftmaxtransformation.to_dict',
                                                                                                                  'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms.IdentityTransformation': ( 'data_utils/transform.html#identitytransformation',
                                                                                                     'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms.IdentityTransformation.__init__': ( 'data_utils/transform.html#identitytransformation.__init__',
                                                                                                              'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms.IdentityTransformation.apply_constraints': ( 'data_utils/transform.html#identitytransformation.apply_constraints',
                                                                                                                       'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms.IdentityTransformation.fit': ( 'data_utils/transform.html#identitytransformation.fit',
                                                                                                         'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms.IdentityTransformation.fit_transform': ( 'data_utils/transform.html#identitytransformation.fit_transform',
                                                                                                                   'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms.IdentityTransformation.from_dict': ( 'data_utils/transform.html#identitytransformation.from_dict',
                                                                                                               'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms.IdentityTransformation.to_dict': ( 'data_utils/transform.html#identitytransformation.to_dict',
                                                                                                             'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms.IdentityTransformation.transform': ( 'data_utils/transform.html#identitytransformation.transform',
                                                                                                               'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms.MinMaxTransformation': ( 'data_utils/transform.html#minmaxtransformation',
                                                                                                   'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms.MinMaxTransformation.__init__': ( 'data_utils/transform.html#minmaxtransformation.__init__',
                                                                                                            'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms.MinMaxTransformation.apply_constraints': ( 'data_utils/transform.html#minmaxtransformation.apply_constraints',
                                                                                                                     'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms.OneHotTransformation': ( 'data_utils/transform.html#onehottransformation',
                                                                                                   'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms.OrdinalTransformation': ( 'data_utils/transform.html#ordinaltransformation',
                                                                                                    'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms.OrdinalTransformation.__init__': ( 'data_utils/transform.html#ordinaltransformation.__init__',
                                                                                                             'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms.OrdinalTransformation.num_categories': ( 'data_utils/transform.html#ordinaltransformation.num_categories',
                                                                                                                   'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms.SoftmaxTransformation': ( 'data_utils/transform.html#softmaxtransformation',
                                                                                                    'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms.SoftmaxTransformation.__init__': ( 'data_utils/transform.html#softmaxtransformation.__init__',
                                                                                                             'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms.SoftmaxTransformation.soft_constraints': ( 'data_utils/transform.html#softmaxtransformation.soft_constraints',
                                                                                                                     'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms._DefaultTransformation': ( 'data_utils/transform.html#_defaulttransformation',
                                                                                                     'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms._DefaultTransformation.apply_constraints': ( 'data_utils/transform.html#_defaulttransformation.apply_constraints',
                                                                                                                       'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms._DefaultTransformation.compute_reg_loss': ( 'data_utils/transform.html#_defaulttransformation.compute_reg_loss',
                                                                                                                      'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms._DefaultTransformation.fit': ( 'data_utils/transform.html#_defaulttransformation.fit',
                                                                                                         'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms._DefaultTransformation.fit_transform': ( 'data_utils/transform.html#_defaulttransformation.fit_transform',
                                                                                                                   'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms._DefaultTransformation.from_dict': ( 'data_utils/transform.html#_defaulttransformation.from_dict',
                                                                                                               'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms._DefaultTransformation.inverse_transform': ( 'data_utils/transform.html#_defaulttransformation.inverse_transform',
                                                                                                                       'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms._DefaultTransformation.is_categorical': ( 'data_utils/transform.html#_defaulttransformation.is_categorical',
                                                                                                                    'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms._DefaultTransformation.to_dict': ( 'data_utils/transform.html#_defaulttransformation.to_dict',
                                                                                                             'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms._DefaultTransformation.transform': ( 'data_utils/transform.html#_defaulttransformation.transform',
                                                                                                               'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms._OneHotTransformation': ( 'data_utils/transform.html#_onehottransformation',
                                                                                                    'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms._OneHotTransformation.__init__': ( 'data_utils/transform.html#_onehottransformation.__init__',
                                                                                                             'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms._OneHotTransformation.apply_constraints': ( 'data_utils/transform.html#_onehottransformation.apply_constraints',
                                                                                                                      'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms._OneHotTransformation.compute_reg_loss': ( 'data_utils/transform.html#_onehottransformation.compute_reg_loss',
                                                                                                                     'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms._OneHotTransformation.hard_constraints': ( 'data_utils/transform.html#_onehottransformation.hard_constraints',
                                                                                                                     'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms._OneHotTransformation.num_categories': ( 'data_utils/transform.html#_onehottransformation.num_categories',
                                                                                                                   'relax/data_utils/transforms.py'),
                                             'relax.data_utils.transforms._OneHotTransformation.soft_constraints': ( 'data_utils/transform.html#_onehottransformation.soft_constraints',
                                                                                                                     'relax/data_utils/transforms.py')},
            'relax.docs': { 'relax.docs.CustomizedMarkdownRenderer': ('docs.html#customizedmarkdownrenderer', 'relax/docs.py'),
                            'relax.docs.CustomizedMarkdownRenderer.__init__': ( 'docs.html#customizedmarkdownrenderer.__init__',
                                                                                'relax/docs.py'),
                            'relax.docs.CustomizedMarkdownRenderer._check_sym': ( 'docs.html#customizedmarkdownrenderer._check_sym',
                                                                                  'relax/docs.py'),
                            'relax.docs.CustomizedMarkdownRenderer._repr_markdown_': ( 'docs.html#customizedmarkdownrenderer._repr_markdown_',
                                                                                       'relax/docs.py'),
                            'relax.docs.CustomizedMarkdownRenderer.hook_methods': ( 'docs.html#customizedmarkdownrenderer.hook_methods',
                                                                                    'relax/docs.py'),
                            'relax.docs.ListDocment': ('docs.html#listdocment', 'relax/docs.py'),
                            'relax.docs.ListDocment.__init__': ('docs.html#listdocment.__init__', 'relax/docs.py'),
                            'relax.docs.ListDocment._repre_mardown': ('docs.html#listdocment._repre_mardown', 'relax/docs.py'),
                            'relax.docs.ParserMarkdownRenderer': ('docs.html#parsermarkdownrenderer', 'relax/docs.py'),
                            'relax.docs.ParserMarkdownRenderer.__init__': ('docs.html#parsermarkdownrenderer.__init__', 'relax/docs.py'),
                            'relax.docs._bold': ('docs.html#_bold', 'relax/docs.py'),
                            'relax.docs._docment_parser': ('docs.html#_docment_parser', 'relax/docs.py'),
                            'relax.docs._fmt_sig': ('docs.html#_fmt_sig', 'relax/docs.py'),
                            'relax.docs._inner_list2mdlist': ('docs.html#_inner_list2mdlist', 'relax/docs.py'),
                            'relax.docs._italic': ('docs.html#_italic', 'relax/docs.py'),
                            'relax.docs._params_mdlist': ('docs.html#_params_mdlist', 'relax/docs.py'),
                            'relax.docs._repr_markdown': ('docs.html#_repr_markdown', 'relax/docs.py'),
                            'relax.docs._return_mdlist': ('docs.html#_return_mdlist', 'relax/docs.py'),
                            'relax.docs._show_param': ('docs.html#_show_param', 'relax/docs.py'),
                            'relax.docs._show_params_return': ('docs.html#_show_params_return', 'relax/docs.py')},
            'relax.evaluate': { 'relax.evaluate.BaseEvalMetrics': ('evaluate.html#baseevalmetrics', 'relax/evaluate.py'),
                                'relax.evaluate.BaseEvalMetrics.__call__': ('evaluate.html#baseevalmetrics.__call__', 'relax/evaluate.py'),
                                'relax.evaluate.BaseEvalMetrics.__init__': ('evaluate.html#baseevalmetrics.__init__', 'relax/evaluate.py'),
                                'relax.evaluate.BaseEvalMetrics.__str__': ('evaluate.html#baseevalmetrics.__str__', 'relax/evaluate.py'),
                                'relax.evaluate.ManifoldDist': ('evaluate.html#manifolddist', 'relax/evaluate.py'),
                                'relax.evaluate.ManifoldDist.__call__': ('evaluate.html#manifolddist.__call__', 'relax/evaluate.py'),
                                'relax.evaluate.ManifoldDist.__init__': ('evaluate.html#manifolddist.__init__', 'relax/evaluate.py'),
                                'relax.evaluate.PredictiveAccuracy': ('evaluate.html#predictiveaccuracy', 'relax/evaluate.py'),
                                'relax.evaluate.PredictiveAccuracy.__call__': ( 'evaluate.html#predictiveaccuracy.__call__',
                                                                                'relax/evaluate.py'),
                                'relax.evaluate.PredictiveAccuracy.__init__': ( 'evaluate.html#predictiveaccuracy.__init__',
                                                                                'relax/evaluate.py'),
                                'relax.evaluate.Proximity': ('evaluate.html#proximity', 'relax/evaluate.py'),
                                'relax.evaluate.Proximity.__call__': ('evaluate.html#proximity.__call__', 'relax/evaluate.py'),
                                'relax.evaluate.Proximity.__init__': ('evaluate.html#proximity.__init__', 'relax/evaluate.py'),
                                'relax.evaluate.Runtime': ('evaluate.html#runtime', 'relax/evaluate.py'),
                                'relax.evaluate.Runtime.__call__': ('evaluate.html#runtime.__call__', 'relax/evaluate.py'),
                                'relax.evaluate.Runtime.__init__': ('evaluate.html#runtime.__init__', 'relax/evaluate.py'),
                                'relax.evaluate.Sparsity': ('evaluate.html#sparsity', 'relax/evaluate.py'),
                                'relax.evaluate.Sparsity.__call__': ('evaluate.html#sparsity.__call__', 'relax/evaluate.py'),
                                'relax.evaluate.Sparsity.__init__': ('evaluate.html#sparsity.__init__', 'relax/evaluate.py'),
                                'relax.evaluate.Validity': ('evaluate.html#validity', 'relax/evaluate.py'),
                                'relax.evaluate.Validity.__call__': ('evaluate.html#validity.__call__', 'relax/evaluate.py'),
                                'relax.evaluate.Validity.__init__': ('evaluate.html#validity.__init__', 'relax/evaluate.py'),
                                'relax.evaluate._get_metric': ('evaluate.html#_get_metric', 'relax/evaluate.py'),
                                'relax.evaluate.benchmark_cfs': ('evaluate.html#benchmark_cfs', 'relax/evaluate.py'),
                                'relax.evaluate.compute_proximity': ('evaluate.html#compute_proximity', 'relax/evaluate.py'),
                                'relax.evaluate.compute_single_proximity': ('evaluate.html#compute_single_proximity', 'relax/evaluate.py'),
                                'relax.evaluate.compute_single_sparsity': ('evaluate.html#compute_single_sparsity', 'relax/evaluate.py'),
                                'relax.evaluate.compute_single_validity': ('evaluate.html#compute_single_validity', 'relax/evaluate.py'),
                                'relax.evaluate.compute_sparsity': ('evaluate.html#compute_sparsity', 'relax/evaluate.py'),
                                'relax.evaluate.compute_validity': ('evaluate.html#compute_validity', 'relax/evaluate.py'),
                                'relax.evaluate.evaluate_cfs': ('evaluate.html#evaluate_cfs', 'relax/evaluate.py'),
                                'relax.evaluate.l2_ann': ('evaluate.html#l2_ann', 'relax/evaluate.py'),
                                'relax.evaluate.pairwise_distances': ('evaluate.html#pairwise_distances', 'relax/evaluate.py')},
            'relax.explain': { 'relax.explain.Explanation': ('explain.html#explanation', 'relax/explain.py'),
                               'relax.explain.Explanation.__getitem__': ('explain.html#explanation.__getitem__', 'relax/explain.py'),
                               'relax.explain.Explanation.__init__': ('explain.html#explanation.__init__', 'relax/explain.py'),
                               'relax.explain.Explanation.__repr__': ('explain.html#explanation.__repr__', 'relax/explain.py'),
                               'relax.explain.Explanation.cfs': ('explain.html#explanation.cfs', 'relax/explain.py'),
                               'relax.explain.Explanation.copy': ('explain.html#explanation.copy', 'relax/explain.py'),
                               'relax.explain.Explanation.data_name': ('explain.html#explanation.data_name', 'relax/explain.py'),
                               'relax.explain.Explanation.feature_indices': ( 'explain.html#explanation.feature_indices',
                                                                              'relax/explain.py'),
                               'relax.explain.Explanation.features_and_indices': ( 'explain.html#explanation.features_and_indices',
                                                                                   'relax/explain.py'),
                               'relax.explain.Explanation.load_from_path': ('explain.html#explanation.load_from_path', 'relax/explain.py'),
                               'relax.explain.Explanation.save': ('explain.html#explanation.save', 'relax/explain.py'),
                               'relax.explain.fake_explanation': ('explain.html#fake_explanation', 'relax/explain.py'),
                               'relax.explain.generate_cf_explanations': ('explain.html#generate_cf_explanations', 'relax/explain.py'),
                               'relax.explain.prepare_cf_module': ('explain.html#prepare_cf_module', 'relax/explain.py'),
                               'relax.explain.prepare_pred_fn': ('explain.html#prepare_pred_fn', 'relax/explain.py'),
                               'relax.explain.prepare_rng_keys': ('explain.html#prepare_rng_keys', 'relax/explain.py')},
            'relax.import_essentials': {},
            'relax.legacy.ckpt_manager': { 'relax.legacy.ckpt_manager.CheckpointManager': ( 'legacy/ckpt_manager.html#checkpointmanager',
                                                                                            'relax/legacy/ckpt_manager.py'),
                                           'relax.legacy.ckpt_manager.CheckpointManager.__init__': ( 'legacy/ckpt_manager.html#checkpointmanager.__init__',
                                                                                                     'relax/legacy/ckpt_manager.py'),
                                           'relax.legacy.ckpt_manager.CheckpointManager.delete_net_opt': ( 'legacy/ckpt_manager.html#checkpointmanager.delete_net_opt',
                                                                                                           'relax/legacy/ckpt_manager.py'),
                                           'relax.legacy.ckpt_manager.CheckpointManager.save_net_opt': ( 'legacy/ckpt_manager.html#checkpointmanager.save_net_opt',
                                                                                                         'relax/legacy/ckpt_manager.py'),
                                           'relax.legacy.ckpt_manager.CheckpointManager.update_checkpoints': ( 'legacy/ckpt_manager.html#checkpointmanager.update_checkpoints',
                                                                                                               'relax/legacy/ckpt_manager.py'),
                                           'relax.legacy.ckpt_manager.load_checkpoint': ( 'legacy/ckpt_manager.html#load_checkpoint',
                                                                                          'relax/legacy/ckpt_manager.py'),
                                           'relax.legacy.ckpt_manager.save_checkpoint': ( 'legacy/ckpt_manager.html#save_checkpoint',
                                                                                          'relax/legacy/ckpt_manager.py')},
            'relax.legacy.import_essentials': {},
            'relax.legacy.logger': { 'relax.legacy.logger.Logger': ('legacy/logger.html#logger', 'relax/legacy/logger.py'),
                                     'relax.legacy.logger.Logger.__init__': ( 'legacy/logger.html#logger.__init__',
                                                                              'relax/legacy/logger.py'),
                                     'relax.legacy.logger.Logger.close': ('legacy/logger.html#logger.close', 'relax/legacy/logger.py'),
                                     'relax.legacy.logger.Logger.get_last_logs': ( 'legacy/logger.html#logger.get_last_logs',
                                                                                   'relax/legacy/logger.py'),
                                     'relax.legacy.logger.Logger.log': ('legacy/logger.html#logger.log', 'relax/legacy/logger.py'),
                                     'relax.legacy.logger.Logger.log_dict': ( 'legacy/logger.html#logger.log_dict',
                                                                              'relax/legacy/logger.py'),
                                     'relax.legacy.logger.Logger.log_dir': ('legacy/logger.html#logger.log_dir', 'relax/legacy/logger.py'),
                                     'relax.legacy.logger.Logger.on_epoch_finished': ( 'legacy/logger.html#logger.on_epoch_finished',
                                                                                       'relax/legacy/logger.py'),
                                     'relax.legacy.logger.Logger.on_epoch_started': ( 'legacy/logger.html#logger.on_epoch_started',
                                                                                      'relax/legacy/logger.py'),
                                     'relax.legacy.logger.Logger.save_hyperparams': ( 'legacy/logger.html#logger.save_hyperparams',
                                                                                      'relax/legacy/logger.py')},
            'relax.legacy.module': { 'relax.legacy.module.BaseNetwork': ('legacy/module.html#basenetwork', 'relax/legacy/module.py'),
                                     'relax.legacy.module.BaseNetwork.__call__': ( 'legacy/module.html#basenetwork.__call__',
                                                                                   'relax/legacy/module.py'),
                                     'relax.legacy.module.BaseTrainingModule': ( 'legacy/module.html#basetrainingmodule',
                                                                                 'relax/legacy/module.py'),
                                     'relax.legacy.module.BaseTrainingModule.init_logger': ( 'legacy/module.html#basetrainingmodule.init_logger',
                                                                                             'relax/legacy/module.py'),
                                     'relax.legacy.module.BaseTrainingModule.init_net_opt': ( 'legacy/module.html#basetrainingmodule.init_net_opt',
                                                                                              'relax/legacy/module.py'),
                                     'relax.legacy.module.BaseTrainingModule.log': ( 'legacy/module.html#basetrainingmodule.log',
                                                                                     'relax/legacy/module.py'),
                                     'relax.legacy.module.BaseTrainingModule.log_dict': ( 'legacy/module.html#basetrainingmodule.log_dict',
                                                                                          'relax/legacy/module.py'),
                                     'relax.legacy.module.BaseTrainingModule.save_hyperparameters': ( 'legacy/module.html#basetrainingmodule.save_hyperparameters',
                                                                                                      'relax/legacy/module.py'),
                                     'relax.legacy.module.BaseTrainingModule.training_step': ( 'legacy/module.html#basetrainingmodule.training_step',
                                                                                               'relax/legacy/module.py'),
                                     'relax.legacy.module.BaseTrainingModule.validation_step': ( 'legacy/module.html#basetrainingmodule.validation_step',
                                                                                                 'relax/legacy/module.py'),
                                     'relax.legacy.module.DenseBlock': ('legacy/module.html#denseblock', 'relax/legacy/module.py'),
                                     'relax.legacy.module.DenseBlock.__call__': ( 'legacy/module.html#denseblock.__call__',
                                                                                  'relax/legacy/module.py'),
                                     'relax.legacy.module.DenseBlock.__init__': ( 'legacy/module.html#denseblock.__init__',
                                                                                  'relax/legacy/module.py'),
                                     'relax.legacy.module.MLP': ('legacy/module.html#mlp', 'relax/legacy/module.py'),
                                     'relax.legacy.module.MLP.__call__': ('legacy/module.html#mlp.__call__', 'relax/legacy/module.py'),
                                     'relax.legacy.module.MLP.__init__': ('legacy/module.html#mlp.__init__', 'relax/legacy/module.py'),
                                     'relax.legacy.module.PredictiveModel': ( 'legacy/module.html#predictivemodel',
                                                                              'relax/legacy/module.py'),
                                     'relax.legacy.module.PredictiveModel.__call__': ( 'legacy/module.html#predictivemodel.__call__',
                                                                                       'relax/legacy/module.py'),
                                     'relax.legacy.module.PredictiveModel.__init__': ( 'legacy/module.html#predictivemodel.__init__',
                                                                                       'relax/legacy/module.py'),
                                     'relax.legacy.module.PredictiveModelConfigs': ( 'legacy/module.html#predictivemodelconfigs',
                                                                                     'relax/legacy/module.py'),
                                     'relax.legacy.module.PredictiveTrainingModule': ( 'legacy/module.html#predictivetrainingmodule',
                                                                                       'relax/legacy/module.py'),
                                     'relax.legacy.module.PredictiveTrainingModule.__init__': ( 'legacy/module.html#predictivetrainingmodule.__init__',
                                                                                                'relax/legacy/module.py'),
                                     'relax.legacy.module.PredictiveTrainingModule._training_step': ( 'legacy/module.html#predictivetrainingmodule._training_step',
                                                                                                      'relax/legacy/module.py'),
                                     'relax.legacy.module.PredictiveTrainingModule.forward': ( 'legacy/module.html#predictivetrainingmodule.forward',
                                                                                               'relax/legacy/module.py'),
                                     'relax.legacy.module.PredictiveTrainingModule.init_net_opt': ( 'legacy/module.html#predictivetrainingmodule.init_net_opt',
                                                                                                    'relax/legacy/module.py'),
                                     'relax.legacy.module.PredictiveTrainingModule.loss_fn': ( 'legacy/module.html#predictivetrainingmodule.loss_fn',
                                                                                               'relax/legacy/module.py'),
                                     'relax.legacy.module.PredictiveTrainingModule.pred_fn': ( 'legacy/module.html#predictivetrainingmodule.pred_fn',
                                                                                               'relax/legacy/module.py'),
                                     'relax.legacy.module.PredictiveTrainingModule.training_step': ( 'legacy/module.html#predictivetrainingmodule.training_step',
                                                                                                     'relax/legacy/module.py'),
                                     'relax.legacy.module.PredictiveTrainingModule.validation_step': ( 'legacy/module.html#predictivetrainingmodule.validation_step',
                                                                                                       'relax/legacy/module.py'),
                                     'relax.legacy.module.PredictiveTrainingModuleConfigs': ( 'legacy/module.html#predictivetrainingmoduleconfigs',
                                                                                              'relax/legacy/module.py')},
            'relax.legacy.trainer': { 'relax.legacy.trainer.TrainingConfigs': ( 'legacy/trainer.html#trainingconfigs',
                                                                                'relax/legacy/trainer.py'),
                                      'relax.legacy.trainer.TrainingConfigs.PRNGSequence': ( 'legacy/trainer.html#trainingconfigs.prngsequence',
                                                                                             'relax/legacy/trainer.py'),
                                      'relax.legacy.trainer.train_model': ('legacy/trainer.html#train_model', 'relax/legacy/trainer.py'),
                                      'relax.legacy.trainer.train_model_with_states': ( 'legacy/trainer.html#train_model_with_states',
                                                                                        'relax/legacy/trainer.py')},
            'relax.legacy.utils': { 'relax.legacy.utils.Config': ('legacy/utils.html#config', 'relax/legacy/utils.py'),
                                    'relax.legacy.utils.Config.default': ('legacy/utils.html#config.default', 'relax/legacy/utils.py'),
                                    'relax.legacy.utils.accuracy': ('legacy/utils.html#accuracy', 'relax/legacy/utils.py'),
                                    'relax.legacy.utils.binary_cross_entropy': ( 'legacy/utils.html#binary_cross_entropy',
                                                                                 'relax/legacy/utils.py'),
                                    'relax.legacy.utils.cat_normalize': ('legacy/utils.html#cat_normalize', 'relax/legacy/utils.py'),
                                    'relax.legacy.utils.check_cat_info': ('legacy/utils.html#check_cat_info', 'relax/legacy/utils.py'),
                                    'relax.legacy.utils.dist': ('legacy/utils.html#dist', 'relax/legacy/utils.py'),
                                    'relax.legacy.utils.get_config': ('legacy/utils.html#get_config', 'relax/legacy/utils.py'),
                                    'relax.legacy.utils.grad_update': ('legacy/utils.html#grad_update', 'relax/legacy/utils.py'),
                                    'relax.legacy.utils.init_net_opt': ('legacy/utils.html#init_net_opt', 'relax/legacy/utils.py'),
                                    'relax.legacy.utils.load_json': ('legacy/utils.html#load_json', 'relax/legacy/utils.py'),
                                    'relax.legacy.utils.make_hk_module': ('legacy/utils.html#make_hk_module', 'relax/legacy/utils.py'),
                                    'relax.legacy.utils.make_model': ('legacy/utils.html#make_model', 'relax/legacy/utils.py'),
                                    'relax.legacy.utils.proximity': ('legacy/utils.html#proximity', 'relax/legacy/utils.py'),
                                    'relax.legacy.utils.sigmoid': ('legacy/utils.html#sigmoid', 'relax/legacy/utils.py'),
                                    'relax.legacy.utils.validate_configs': ('legacy/utils.html#validate_configs', 'relax/legacy/utils.py')},
            'relax.methods.base': { 'relax.methods.base.CFModule': ('methods/base.html#cfmodule', 'relax/methods/base.py'),
                                    'relax.methods.base.CFModule.__init__': ( 'methods/base.html#cfmodule.__init__',
                                                                              'relax/methods/base.py'),
                                    'relax.methods.base.CFModule.apply_constraints': ( 'methods/base.html#cfmodule.apply_constraints',
                                                                                       'relax/methods/base.py'),
                                    'relax.methods.base.CFModule.before_generate_cf': ( 'methods/base.html#cfmodule.before_generate_cf',
                                                                                        'relax/methods/base.py'),
                                    'relax.methods.base.CFModule.compute_reg_loss': ( 'methods/base.html#cfmodule.compute_reg_loss',
                                                                                      'relax/methods/base.py'),
                                    'relax.methods.base.CFModule.generate_cf': ( 'methods/base.html#cfmodule.generate_cf',
                                                                                 'relax/methods/base.py'),
                                    'relax.methods.base.CFModule.set_apply_constraints_fn': ( 'methods/base.html#cfmodule.set_apply_constraints_fn',
                                                                                              'relax/methods/base.py'),
                                    'relax.methods.base.CFModule.set_compute_reg_loss_fn': ( 'methods/base.html#cfmodule.set_compute_reg_loss_fn',
                                                                                             'relax/methods/base.py'),
                                    'relax.methods.base.CFModule.set_data_module': ( 'methods/base.html#cfmodule.set_data_module',
                                                                                     'relax/methods/base.py'),
                                    'relax.methods.base.ParametricCFModule': ( 'methods/base.html#parametriccfmodule',
                                                                               'relax/methods/base.py'),
                                    'relax.methods.base.ParametricCFModule.train': ( 'methods/base.html#parametriccfmodule.train',
                                                                                     'relax/methods/base.py'),
                                    'relax.methods.base.default_apply_constraints_fn': ( 'methods/base.html#default_apply_constraints_fn',
                                                                                         'relax/methods/base.py'),
                                    'relax.methods.base.default_compute_reg_loss_fn': ( 'methods/base.html#default_compute_reg_loss_fn',
                                                                                        'relax/methods/base.py')},
            'relax.methods.cchvae': { 'relax.methods.cchvae.CCHVAE': ('methods/cchvae.html#cchvae', 'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae.CCHVAE.__init__': ( 'methods/cchvae.html#cchvae.__init__',
                                                                                'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae.CCHVAE._init_model': ( 'methods/cchvae.html#cchvae._init_model',
                                                                                   'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae.CCHVAE.generate_cf': ( 'methods/cchvae.html#cchvae.generate_cf',
                                                                                   'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae.CCHVAE.train': ('methods/cchvae.html#cchvae.train', 'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae.CCHVAEConfig': ('methods/cchvae.html#cchvaeconfig', 'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae.CHVAE': ('methods/cchvae.html#chvae', 'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae.CHVAE.__init__': ( 'methods/cchvae.html#chvae.__init__',
                                                                               'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae.CHVAE.build': ('methods/cchvae.html#chvae.build', 'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae.CHVAE.call': ('methods/cchvae.html#chvae.call', 'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae.CHVAE.compute_kl_loss': ( 'methods/cchvae.html#chvae.compute_kl_loss',
                                                                                      'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae.CHVAE.decode': ('methods/cchvae.html#chvae.decode', 'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae.CHVAE.encode': ('methods/cchvae.html#chvae.encode', 'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae.CHVAE.forward': ( 'methods/cchvae.html#chvae.forward',
                                                                              'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae.CHVAE.regenerate': ( 'methods/cchvae.html#chvae.regenerate',
                                                                                 'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae.CHVAE.reparameterize': ( 'methods/cchvae.html#chvae.reparameterize',
                                                                                     'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae.CHVAE.set_apply_constraints_fn': ( 'methods/cchvae.html#chvae.set_apply_constraints_fn',
                                                                                               'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae._cchvae': ('methods/cchvae.html#_cchvae', 'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae._hyper_sphere_coordindates': ( 'methods/cchvae.html#_hyper_sphere_coordindates',
                                                                                           'relax/methods/cchvae.py')},
            'relax.methods.clue': { 'relax.methods.clue.CLUE': ('methods/clue.html#clue', 'relax/methods/clue.py'),
                                    'relax.methods.clue.CLUE.__init__': ('methods/clue.html#clue.__init__', 'relax/methods/clue.py'),
                                    'relax.methods.clue.CLUE._init_model': ('methods/clue.html#clue._init_model', 'relax/methods/clue.py'),
                                    'relax.methods.clue.CLUE.generate_cf': ('methods/clue.html#clue.generate_cf', 'relax/methods/clue.py'),
                                    'relax.methods.clue.CLUE.train': ('methods/clue.html#clue.train', 'relax/methods/clue.py'),
                                    'relax.methods.clue.CLUEConfig': ('methods/clue.html#clueconfig', 'relax/methods/clue.py'),
                                    'relax.methods.clue.Decoder': ('methods/clue.html#decoder', 'relax/methods/clue.py'),
                                    'relax.methods.clue.Decoder.__call__': ('methods/clue.html#decoder.__call__', 'relax/methods/clue.py'),
                                    'relax.methods.clue.Decoder.__init__': ('methods/clue.html#decoder.__init__', 'relax/methods/clue.py'),
                                    'relax.methods.clue.Encoder': ('methods/clue.html#encoder', 'relax/methods/clue.py'),
                                    'relax.methods.clue.Encoder.__init__': ('methods/clue.html#encoder.__init__', 'relax/methods/clue.py'),
                                    'relax.methods.clue.Encoder.call': ('methods/clue.html#encoder.call', 'relax/methods/clue.py'),
                                    'relax.methods.clue.VAEGaussCat': ('methods/clue.html#vaegausscat', 'relax/methods/clue.py'),
                                    'relax.methods.clue.VAEGaussCat.__init__': ( 'methods/clue.html#vaegausscat.__init__',
                                                                                 'relax/methods/clue.py'),
                                    'relax.methods.clue.VAEGaussCat.build': ( 'methods/clue.html#vaegausscat.build',
                                                                              'relax/methods/clue.py'),
                                    'relax.methods.clue.VAEGaussCat.call': ('methods/clue.html#vaegausscat.call', 'relax/methods/clue.py'),
                                    'relax.methods.clue.VAEGaussCat.compute_vae_loss': ( 'methods/clue.html#vaegausscat.compute_vae_loss',
                                                                                         'relax/methods/clue.py'),
                                    'relax.methods.clue.VAEGaussCat.decode': ( 'methods/clue.html#vaegausscat.decode',
                                                                               'relax/methods/clue.py'),
                                    'relax.methods.clue.VAEGaussCat.encode': ( 'methods/clue.html#vaegausscat.encode',
                                                                               'relax/methods/clue.py'),
                                    'relax.methods.clue.VAEGaussCat.sample': ( 'methods/clue.html#vaegausscat.sample',
                                                                               'relax/methods/clue.py'),
                                    'relax.methods.clue.VAEGaussCat.sample_latent': ( 'methods/clue.html#vaegausscat.sample_latent',
                                                                                      'relax/methods/clue.py'),
                                    'relax.methods.clue.VAEGaussCat.sample_step': ( 'methods/clue.html#vaegausscat.sample_step',
                                                                                    'relax/methods/clue.py'),
                                    'relax.methods.clue.VAEGaussCat.set_reconstruction_loss': ( 'methods/clue.html#vaegausscat.set_reconstruction_loss',
                                                                                                'relax/methods/clue.py'),
                                    'relax.methods.clue._clue_generate': ('methods/clue.html#_clue_generate', 'relax/methods/clue.py'),
                                    'relax.methods.clue.get_reconstruction_loss_fn': ( 'methods/clue.html#get_reconstruction_loss_fn',
                                                                                       'relax/methods/clue.py'),
                                    'relax.methods.clue.kl_divergence': ('methods/clue.html#kl_divergence', 'relax/methods/clue.py')},
            'relax.methods.counternet': { 'relax.methods.counternet.CounterNet': ( 'methods/counternet.html#counternet',
                                                                                   'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNet.__init__': ( 'methods/counternet.html#counternet.__init__',
                                                                                            'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNet._init_model': ( 'methods/counternet.html#counternet._init_model',
                                                                                               'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNet.generate_cf': ( 'methods/counternet.html#counternet.generate_cf',
                                                                                               'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNet.is_trained': ( 'methods/counternet.html#counternet.is_trained',
                                                                                              'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNet.pred_fn': ( 'methods/counternet.html#counternet.pred_fn',
                                                                                           'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNet.train': ( 'methods/counternet.html#counternet.train',
                                                                                         'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetConfig': ( 'methods/counternet.html#counternetconfig',
                                                                                         'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetModel': ( 'methods/counternet.html#counternetmodel',
                                                                                        'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetModel.__call__': ( 'methods/counternet.html#counternetmodel.__call__',
                                                                                                 'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetModel.__init__': ( 'methods/counternet.html#counternetmodel.__init__',
                                                                                                 'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetTrainingModule': ( 'methods/counternet.html#counternettrainingmodule',
                                                                                                 'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetTrainingModule.__init__': ( 'methods/counternet.html#counternettrainingmodule.__init__',
                                                                                                          'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetTrainingModule._explainer_step': ( 'methods/counternet.html#counternettrainingmodule._explainer_step',
                                                                                                                 'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetTrainingModule._predictor_step': ( 'methods/counternet.html#counternettrainingmodule._predictor_step',
                                                                                                                 'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetTrainingModule._training_step': ( 'methods/counternet.html#counternettrainingmodule._training_step',
                                                                                                                'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetTrainingModule._training_step_logs': ( 'methods/counternet.html#counternettrainingmodule._training_step_logs',
                                                                                                                     'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetTrainingModule.exp_loss_fn': ( 'methods/counternet.html#counternettrainingmodule.exp_loss_fn',
                                                                                                             'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetTrainingModule.forward': ( 'methods/counternet.html#counternettrainingmodule.forward',
                                                                                                         'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetTrainingModule.generate_cf': ( 'methods/counternet.html#counternettrainingmodule.generate_cf',
                                                                                                             'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetTrainingModule.init_net_opt': ( 'methods/counternet.html#counternettrainingmodule.init_net_opt',
                                                                                                              'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetTrainingModule.loss_fn_1': ( 'methods/counternet.html#counternettrainingmodule.loss_fn_1',
                                                                                                           'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetTrainingModule.loss_fn_2': ( 'methods/counternet.html#counternettrainingmodule.loss_fn_2',
                                                                                                           'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetTrainingModule.loss_fn_3': ( 'methods/counternet.html#counternettrainingmodule.loss_fn_3',
                                                                                                           'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetTrainingModule.pred_fn': ( 'methods/counternet.html#counternettrainingmodule.pred_fn',
                                                                                                         'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetTrainingModule.pred_loss_fn': ( 'methods/counternet.html#counternettrainingmodule.pred_loss_fn',
                                                                                                              'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetTrainingModule.training_step': ( 'methods/counternet.html#counternettrainingmodule.training_step',
                                                                                                               'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetTrainingModule.validation_step': ( 'methods/counternet.html#counternettrainingmodule.validation_step',
                                                                                                                 'relax/methods/counternet.py'),
                                          'relax.methods.counternet.partition_trainable_params': ( 'methods/counternet.html#partition_trainable_params',
                                                                                                   'relax/methods/counternet.py')},
            'relax.methods.dice': { 'relax.methods.dice.DiverseCF': ('methods/dice.html#diversecf', 'relax/methods/dice.py'),
                                    'relax.methods.dice.DiverseCF.__init__': ( 'methods/dice.html#diversecf.__init__',
                                                                               'relax/methods/dice.py'),
                                    'relax.methods.dice.DiverseCF.generate_cf': ( 'methods/dice.html#diversecf.generate_cf',
                                                                                  'relax/methods/dice.py'),
                                    'relax.methods.dice.DiverseCF.load_from_path': ( 'methods/dice.html#diversecf.load_from_path',
                                                                                     'relax/methods/dice.py'),
                                    'relax.methods.dice.DiverseCF.save': ('methods/dice.html#diversecf.save', 'relax/methods/dice.py'),
                                    'relax.methods.dice.DiverseCFConfig': ('methods/dice.html#diversecfconfig', 'relax/methods/dice.py'),
                                    'relax.methods.dice._diverse_cf': ('methods/dice.html#_diverse_cf', 'relax/methods/dice.py'),
                                    'relax.methods.dice.dpp_style_vmap': ('methods/dice.html#dpp_style_vmap', 'relax/methods/dice.py')},
            'relax.methods.l2c': { 'relax.methods.l2c.Discretizer': ('methods/l2c.html#discretizer', 'relax/methods/l2c.py'),
                                   'relax.methods.l2c.Discretizer.__init__': ( 'methods/l2c.html#discretizer.__init__',
                                                                               'relax/methods/l2c.py'),
                                   'relax.methods.l2c.Discretizer.fit': ('methods/l2c.html#discretizer.fit', 'relax/methods/l2c.py'),
                                   'relax.methods.l2c.Discretizer.fit_transform': ( 'methods/l2c.html#discretizer.fit_transform',
                                                                                    'relax/methods/l2c.py'),
                                   'relax.methods.l2c.Discretizer.get_pred_fn': ( 'methods/l2c.html#discretizer.get_pred_fn',
                                                                                  'relax/methods/l2c.py'),
                                   'relax.methods.l2c.Discretizer.inverse_transform': ( 'methods/l2c.html#discretizer.inverse_transform',
                                                                                        'relax/methods/l2c.py'),
                                   'relax.methods.l2c.Discretizer.inverse_transform_indices': ( 'methods/l2c.html#discretizer.inverse_transform_indices',
                                                                                                'relax/methods/l2c.py'),
                                   'relax.methods.l2c.Discretizer.inversed_transform_pytree': ( 'methods/l2c.html#discretizer.inversed_transform_pytree',
                                                                                                'relax/methods/l2c.py'),
                                   'relax.methods.l2c.Discretizer.transform': ( 'methods/l2c.html#discretizer.transform',
                                                                                'relax/methods/l2c.py'),
                                   'relax.methods.l2c.Discretizer.transform_indices': ( 'methods/l2c.html#discretizer.transform_indices',
                                                                                        'relax/methods/l2c.py'),
                                   'relax.methods.l2c.L2C': ('methods/l2c.html#l2c', 'relax/methods/l2c.py'),
                                   'relax.methods.l2c.L2C.__init__': ('methods/l2c.html#l2c.__init__', 'relax/methods/l2c.py'),
                                   'relax.methods.l2c.L2C.generate_cf': ('methods/l2c.html#l2c.generate_cf', 'relax/methods/l2c.py'),
                                   'relax.methods.l2c.L2C.train': ('methods/l2c.html#l2c.train', 'relax/methods/l2c.py'),
                                   'relax.methods.l2c.L2CConfig': ('methods/l2c.html#l2cconfig', 'relax/methods/l2c.py'),
                                   'relax.methods.l2c.L2CModel': ('methods/l2c.html#l2cmodel', 'relax/methods/l2c.py'),
                                   'relax.methods.l2c.L2CModel.__init__': ('methods/l2c.html#l2cmodel.__init__', 'relax/methods/l2c.py'),
                                   'relax.methods.l2c.L2CModel.build': ('methods/l2c.html#l2cmodel.build', 'relax/methods/l2c.py'),
                                   'relax.methods.l2c.L2CModel.call': ('methods/l2c.html#l2cmodel.call', 'relax/methods/l2c.py'),
                                   'relax.methods.l2c.L2CModel.compute_l2c_loss': ( 'methods/l2c.html#l2cmodel.compute_l2c_loss',
                                                                                    'relax/methods/l2c.py'),
                                   'relax.methods.l2c.L2CModel.forward': ('methods/l2c.html#l2cmodel.forward', 'relax/methods/l2c.py'),
                                   'relax.methods.l2c.L2CModel.set_features_info': ( 'methods/l2c.html#l2cmodel.set_features_info',
                                                                                     'relax/methods/l2c.py'),
                                   'relax.methods.l2c.L2CModel.set_immutable_mask': ( 'methods/l2c.html#l2cmodel.set_immutable_mask',
                                                                                      'relax/methods/l2c.py'),
                                   'relax.methods.l2c.L2CModel.set_pred_fn': ( 'methods/l2c.html#l2cmodel.set_pred_fn',
                                                                               'relax/methods/l2c.py'),
                                   'relax.methods.l2c.L2CModel.start_end_indices': ( 'methods/l2c.html#l2cmodel.start_end_indices',
                                                                                     'relax/methods/l2c.py'),
                                   'relax.methods.l2c.cut_quantiles': ('methods/l2c.html#cut_quantiles', 'relax/methods/l2c.py'),
                                   'relax.methods.l2c.discretize_xs': ('methods/l2c.html#discretize_xs', 'relax/methods/l2c.py'),
                                   'relax.methods.l2c.gumbel_softmax': ('methods/l2c.html#gumbel_softmax', 'relax/methods/l2c.py'),
                                   'relax.methods.l2c.qcut': ('methods/l2c.html#qcut', 'relax/methods/l2c.py'),
                                   'relax.methods.l2c.qcut_inverse': ('methods/l2c.html#qcut_inverse', 'relax/methods/l2c.py'),
                                   'relax.methods.l2c.sample_bernouli': ('methods/l2c.html#sample_bernouli', 'relax/methods/l2c.py'),
                                   'relax.methods.l2c.sample_categorical': ('methods/l2c.html#sample_categorical', 'relax/methods/l2c.py'),
                                   'relax.methods.l2c.split_fn': ('methods/l2c.html#split_fn', 'relax/methods/l2c.py')},
            'relax.methods.proto': { 'relax.methods.proto.ProtoCF': ('methods/proto.html#protocf', 'relax/methods/proto.py'),
                                     'relax.methods.proto.ProtoCF.__init__': ( 'methods/proto.html#protocf.__init__',
                                                                               'relax/methods/proto.py'),
                                     'relax.methods.proto.ProtoCF._init_model': ( 'methods/proto.html#protocf._init_model',
                                                                                  'relax/methods/proto.py'),
                                     'relax.methods.proto.ProtoCF.generate_cf': ( 'methods/proto.html#protocf.generate_cf',
                                                                                  'relax/methods/proto.py'),
                                     'relax.methods.proto.ProtoCF.train': ('methods/proto.html#protocf.train', 'relax/methods/proto.py'),
                                     'relax.methods.proto.ProtoCFConfig': ('methods/proto.html#protocfconfig', 'relax/methods/proto.py'),
                                     'relax.methods.proto._proto_cf': ('methods/proto.html#_proto_cf', 'relax/methods/proto.py')},
            'relax.methods.sphere': { 'relax.methods.sphere.GSConfig': ('methods/sphere.html#gsconfig', 'relax/methods/sphere.py'),
                                      'relax.methods.sphere.GrowingSphere': ( 'methods/sphere.html#growingsphere',
                                                                              'relax/methods/sphere.py'),
                                      'relax.methods.sphere.GrowingSphere.__init__': ( 'methods/sphere.html#growingsphere.__init__',
                                                                                       'relax/methods/sphere.py'),
                                      'relax.methods.sphere.GrowingSphere.before_generate_cf': ( 'methods/sphere.html#growingsphere.before_generate_cf',
                                                                                                 'relax/methods/sphere.py'),
                                      'relax.methods.sphere.GrowingSphere.generate_cf': ( 'methods/sphere.html#growingsphere.generate_cf',
                                                                                          'relax/methods/sphere.py'),
                                      'relax.methods.sphere.GrowingSphere.has_data_module': ( 'methods/sphere.html#growingsphere.has_data_module',
                                                                                              'relax/methods/sphere.py'),
                                      'relax.methods.sphere.GrowingSphere.load_from_path': ( 'methods/sphere.html#growingsphere.load_from_path',
                                                                                             'relax/methods/sphere.py'),
                                      'relax.methods.sphere.GrowingSphere.save': ( 'methods/sphere.html#growingsphere.save',
                                                                                   'relax/methods/sphere.py'),
                                      'relax.methods.sphere._growing_spheres': ( 'methods/sphere.html#_growing_spheres',
                                                                                 'relax/methods/sphere.py'),
                                      'relax.methods.sphere.cat_perturb_fn': ( 'methods/sphere.html#cat_perturb_fn',
                                                                               'relax/methods/sphere.py'),
                                      'relax.methods.sphere.default_perturb_function': ( 'methods/sphere.html#default_perturb_function',
                                                                                         'relax/methods/sphere.py'),
                                      'relax.methods.sphere.features_to_infos_and_perturb_fn': ( 'methods/sphere.html#features_to_infos_and_perturb_fn',
                                                                                                 'relax/methods/sphere.py'),
                                      'relax.methods.sphere.hyper_sphere_coordindates': ( 'methods/sphere.html#hyper_sphere_coordindates',
                                                                                          'relax/methods/sphere.py'),
                                      'relax.methods.sphere.perturb_function_with_features': ( 'methods/sphere.html#perturb_function_with_features',
                                                                                               'relax/methods/sphere.py'),
                                      'relax.methods.sphere.sample_categorical': ( 'methods/sphere.html#sample_categorical',
                                                                                   'relax/methods/sphere.py')},
            'relax.methods.vaecf': { 'relax.methods.vaecf.VAE': ('methods/vaecf.html#vae', 'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.VAE.__init__': ('methods/vaecf.html#vae.__init__', 'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.VAE._compile': ('methods/vaecf.html#vae._compile', 'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.VAE.build': ('methods/vaecf.html#vae.build', 'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.VAE.call': ('methods/vaecf.html#vae.call', 'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.VAE.compute_vae_loss': ( 'methods/vaecf.html#vae.compute_vae_loss',
                                                                                   'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.VAE.decode': ('methods/vaecf.html#vae.decode', 'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.VAE.encode': ('methods/vaecf.html#vae.encode', 'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.VAE.sample': ('methods/vaecf.html#vae.sample', 'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.VAE.set_compute_regularization_fn': ( 'methods/vaecf.html#vae.set_compute_regularization_fn',
                                                                                                'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.VAE.set_pred_fn': ( 'methods/vaecf.html#vae.set_pred_fn',
                                                                              'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.VAECF': ('methods/vaecf.html#vaecf', 'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.VAECF.__init__': ('methods/vaecf.html#vaecf.__init__', 'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.VAECF._init_model': ( 'methods/vaecf.html#vaecf._init_model',
                                                                                'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.VAECF.generate_cf': ( 'methods/vaecf.html#vaecf.generate_cf',
                                                                                'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.VAECF.train': ('methods/vaecf.html#vaecf.train', 'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.VAECFConfig': ('methods/vaecf.html#vaecfconfig', 'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.hindge_embedding_loss': ( 'methods/vaecf.html#hindge_embedding_loss',
                                                                                    'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.sample_latent': ('methods/vaecf.html#sample_latent', 'relax/methods/vaecf.py')},
            'relax.methods.vanilla': { 'relax.methods.vanilla.VanillaCF': ('methods/vanilla.html#vanillacf', 'relax/methods/vanilla.py'),
                                       'relax.methods.vanilla.VanillaCF.__init__': ( 'methods/vanilla.html#vanillacf.__init__',
                                                                                     'relax/methods/vanilla.py'),
                                       'relax.methods.vanilla.VanillaCF.generate_cf': ( 'methods/vanilla.html#vanillacf.generate_cf',
                                                                                        'relax/methods/vanilla.py'),
                                       'relax.methods.vanilla.VanillaCF.load_from_path': ( 'methods/vanilla.html#vanillacf.load_from_path',
                                                                                           'relax/methods/vanilla.py'),
                                       'relax.methods.vanilla.VanillaCF.save': ( 'methods/vanilla.html#vanillacf.save',
                                                                                 'relax/methods/vanilla.py'),
                                       'relax.methods.vanilla.VanillaCFConfig': ( 'methods/vanilla.html#vanillacfconfig',
                                                                                  'relax/methods/vanilla.py'),
                                       'relax.methods.vanilla._vanilla_cf': ( 'methods/vanilla.html#_vanilla_cf',
                                                                              'relax/methods/vanilla.py')},
            'relax.ml_model': { 'relax.ml_model.AutoEncoder': ('ml_model.html#autoencoder', 'relax/ml_model.py'),
                                'relax.ml_model.AutoEncoder.__init__': ('ml_model.html#autoencoder.__init__', 'relax/ml_model.py'),
                                'relax.ml_model.AutoEncoder.call': ('ml_model.html#autoencoder.call', 'relax/ml_model.py'),
                                'relax.ml_model.MLModule': ('ml_model.html#mlmodule', 'relax/ml_model.py'),
                                'relax.ml_model.MLModule.__init__': ('ml_model.html#mlmodule.__init__', 'relax/ml_model.py'),
                                'relax.ml_model.MLModule._init_model': ('ml_model.html#mlmodule._init_model', 'relax/ml_model.py'),
                                'relax.ml_model.MLModule.is_trained': ('ml_model.html#mlmodule.is_trained', 'relax/ml_model.py'),
                                'relax.ml_model.MLModule.load_from_path': ('ml_model.html#mlmodule.load_from_path', 'relax/ml_model.py'),
                                'relax.ml_model.MLModule.pred_fn': ('ml_model.html#mlmodule.pred_fn', 'relax/ml_model.py'),
                                'relax.ml_model.MLModule.save': ('ml_model.html#mlmodule.save', 'relax/ml_model.py'),
                                'relax.ml_model.MLModule.train': ('ml_model.html#mlmodule.train', 'relax/ml_model.py'),
                                'relax.ml_model.MLModuleConfig': ('ml_model.html#mlmoduleconfig', 'relax/ml_model.py'),
                                'relax.ml_model.MLP': ('ml_model.html#mlp', 'relax/ml_model.py'),
                                'relax.ml_model.MLP.__init__': ('ml_model.html#mlp.__init__', 'relax/ml_model.py'),
                                'relax.ml_model.MLP.call': ('ml_model.html#mlp.call', 'relax/ml_model.py'),
                                'relax.ml_model.MLP.get_config': ('ml_model.html#mlp.get_config', 'relax/ml_model.py'),
                                'relax.ml_model.MLPBlock': ('ml_model.html#mlpblock', 'relax/ml_model.py'),
                                'relax.ml_model.MLPBlock.__init__': ('ml_model.html#mlpblock.__init__', 'relax/ml_model.py'),
                                'relax.ml_model.MLPBlock.build': ('ml_model.html#mlpblock.build', 'relax/ml_model.py'),
                                'relax.ml_model.MLPBlock.call': ('ml_model.html#mlpblock.call', 'relax/ml_model.py'),
                                'relax.ml_model.download_ml_module': ('ml_model.html#download_ml_module', 'relax/ml_model.py'),
                                'relax.ml_model.load_ml_module': ('ml_model.html#load_ml_module', 'relax/ml_model.py')},
            'relax.strategy': { 'relax.strategy.BaseStrategy': ('explain.strategy.html#basestrategy', 'relax/strategy.py'),
                                'relax.strategy.BaseStrategy.__call__': ( 'explain.strategy.html#basestrategy.__call__',
                                                                          'relax/strategy.py'),
                                'relax.strategy.BatchedPmapStrategy': ('explain.strategy.html#batchedpmapstrategy', 'relax/strategy.py'),
                                'relax.strategy.BatchedPmapStrategy.__call__': ( 'explain.strategy.html#batchedpmapstrategy.__call__',
                                                                                 'relax/strategy.py'),
                                'relax.strategy.BatchedPmapStrategy.__init__': ( 'explain.strategy.html#batchedpmapstrategy.__init__',
                                                                                 'relax/strategy.py'),
                                'relax.strategy.BatchedVmapStrategy': ('explain.strategy.html#batchedvmapstrategy', 'relax/strategy.py'),
                                'relax.strategy.BatchedVmapStrategy.__call__': ( 'explain.strategy.html#batchedvmapstrategy.__call__',
                                                                                 'relax/strategy.py'),
                                'relax.strategy.BatchedVmapStrategy.__init__': ( 'explain.strategy.html#batchedvmapstrategy.__init__',
                                                                                 'relax/strategy.py'),
                                'relax.strategy.IterativeStrategy': ('explain.strategy.html#iterativestrategy', 'relax/strategy.py'),
                                'relax.strategy.IterativeStrategy.__call__': ( 'explain.strategy.html#iterativestrategy.__call__',
                                                                               'relax/strategy.py'),
                                'relax.strategy.PmapStrategy': ('explain.strategy.html#pmapstrategy', 'relax/strategy.py'),
                                'relax.strategy.PmapStrategy.__call__': ( 'explain.strategy.html#pmapstrategy.__call__',
                                                                          'relax/strategy.py'),
                                'relax.strategy.PmapStrategy.__init__': ( 'explain.strategy.html#pmapstrategy.__init__',
                                                                          'relax/strategy.py'),
                                'relax.strategy.StrategyFactory': ('explain.strategy.html#strategyfactory', 'relax/strategy.py'),
                                'relax.strategy.StrategyFactory.__init__': ( 'explain.strategy.html#strategyfactory.__init__',
                                                                             'relax/strategy.py'),
                                'relax.strategy.StrategyFactory.get_default_strategy': ( 'explain.strategy.html#strategyfactory.get_default_strategy',
                                                                                         'relax/strategy.py'),
                                'relax.strategy.StrategyFactory.get_strategy': ( 'explain.strategy.html#strategyfactory.get_strategy',
                                                                                 'relax/strategy.py'),
                                'relax.strategy.VmapStrategy': ('explain.strategy.html#vmapstrategy', 'relax/strategy.py'),
                                'relax.strategy.VmapStrategy.__call__': ( 'explain.strategy.html#vmapstrategy.__call__',
                                                                          'relax/strategy.py'),
                                'relax.strategy._batched_generation': ('explain.strategy.html#_batched_generation', 'relax/strategy.py'),
                                'relax.strategy._pad_divisible_X': ('explain.strategy.html#_pad_divisible_x', 'relax/strategy.py'),
                                'relax.strategy._pad_xs': ('explain.strategy.html#_pad_xs', 'relax/strategy.py')},
            'relax.utils': { 'relax.utils.Config': ('utils.html#config', 'relax/utils.py'),
                             'relax.utils.Config.default': ('utils.html#config.default', 'relax/utils.py'),
                             'relax.utils._is_array': ('utils.html#_is_array', 'relax/utils.py'),
                             'relax.utils._reshape_x': ('utils.html#_reshape_x', 'relax/utils.py'),
                             'relax.utils.auto_reshaping': ('utils.html#auto_reshaping', 'relax/utils.py'),
                             'relax.utils.get_config': ('utils.html#get_config', 'relax/utils.py'),
                             'relax.utils.grad_update': ('utils.html#grad_update', 'relax/utils.py'),
                             'relax.utils.gumbel_softmax': ('utils.html#gumbel_softmax', 'relax/utils.py'),
                             'relax.utils.load_json': ('utils.html#load_json', 'relax/utils.py'),
                             'relax.utils.load_pytree': ('utils.html#load_pytree', 'relax/utils.py'),
                             'relax.utils.save_pytree': ('utils.html#save_pytree', 'relax/utils.py'),
                             'relax.utils.set_config': ('utils.html#set_config', 'relax/utils.py'),
                             'relax.utils.validate_configs': ('utils.html#validate_configs', 'relax/utils.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/methods/05_sphere.ipynb.

# %% ../../nbs/methods/05_sphere.ipynb 3
from __future__ import annotations
from ..import_essentials import *
from .base import CFModule, BaseConfig, default_apply_constraints_fn
from ..utils import auto_reshaping, grad_update, validate_configs
from ..data_utils import Feature, FeaturesList
from ..data_module import DataModule

# %% auto 0
__all__ = ['hyper_sphere_coordindates', 'sample_categorical', 'default_perturb_function', 'perturb_function_with_features',
           'GSConfig', 'GrowingSphere']

# %% ../../nbs/methods/05_sphere.ipynb 5
@partial(jit, static_argnums=(2, 5))
def hyper_sphere_coordindates(
    rng_key: jrand.PRNGKey, # Random number generator key
    x: Array, # Input instance with only continuous features. Shape: (1, n_features)
    n_samples: int, # Number of samples
    high: float, # Upper bound
    low: float, # Lower bound
    p_norm: int = 2 # Norm
):
    # Adapted from 
    # https://github.com/carla-recourse/CARLA/blob/24db00aa8616eb2faedea0d6edf6e307cee9d192/carla/recourse_methods/catalog/growing_spheres/library/gs_counterfactuals.py#L8
    key_1, key_2 = jrand.split(rng_key)
    delta = jrand.normal(key_1, shape=(n_samples, x.shape[-1]))
    dist = jrand.uniform(key_2, shape=(n_samples,)) * (high - low) + low
    norm_p = jnp.linalg.norm(delta, ord=p_norm, axis=1)
    d_norm = jnp.divide(dist, norm_p).reshape(-1, 1)  # rescale/normalize factor
    delta = jnp.multiply(delta, d_norm)
    candidates = x + delta

    return candidates

# %% ../../nbs/methods/05_sphere.ipynb 6
def sample_categorical(rng_key: jrand.PRNGKey, col_size: int, n_samples: int):
    rng_key, _ = jrand.split(rng_key)
    prob = jnp.ones(col_size) / col_size
    cat_sample = jrand.categorical(rng_key, prob, shape=(n_samples, 1))
    return cat_sample

# %% ../../nbs/methods/05_sphere.ipynb 7
def default_perturb_function(
    rng_key: jrand.PRNGKey,
    x: np.ndarray, # Shape: (1, k)
    n_samples: int,
    high: float,
    low: float,
    p_norm: int
):
    return hyper_sphere_coordindates(
        rng_key, x, n_samples, high, low, p_norm
    )

# def perturb_function_with_features(
#     rng_key: jrand.PRNGKey,
#     x: np.ndarray, # Shape: (1, k)
#     n_samples: int,
#     high, 
#     low,
#     p_norm,
#     feats: FeaturesList,
# ):
#     def perturb_feature(rng_key, x, feat):
#         if feat.is_categorical:
#             sampled_cat = sample_categorical(
#                 rng_key, feat.transformation.num_categories, n_samples
#             ) #<== sampled labels
#             transformation = feat.transformation.name
#             if transformation == 'ohe':
#                 return jax.nn.one_hot(
#                     sampled_cat.reshape(-1), num_classes=feat.transformation.num_categories
#                 ) #<== transformed labels
#             elif transformation == 'ordinal':
#                 return sampled_cat
#             else:
#                 raise NotImplementedError
#         else: 
#             return hyper_sphere_coordindates(
#                 rng_key, x, n_samples, high, low, p_norm
#             ) #<== transformed continuous features
        
#     rng_keys = jrand.split(rng_key, len(feats))
#     perturbed = jnp.repeat(x, n_samples, axis=0)
#     for rng_key, (start, end), feat in zip(rng_keys, feats.feature_indices, feats):
#         _perturbed_feat = perturb_feature(rng_keys[0], x[:, start: end], feat)
#         perturbed = perturbed.at[:, start: end].set(_perturbed_feat)
#     return perturbed


# %% ../../nbs/methods/05_sphere.ipynb 8
@partial(jit, static_argnums=(2, 5, 8, 9))
def perturb_function_with_features(
    rng_key: jrand.PRNGKey,
    x: np.ndarray, # Shape: (1, k)
    n_samples: int,
    high: float, 
    low: float,
    p_norm: int,
    cont_masks: Array,
    immut_masks: Array,
    num_categories: list[int],
    cat_perturb_fn: Callable
):
        
    def perturb_cat_feat(rng_key, num_categories):
        rng_key, next_key = jrand.split(rng_key)
        sampled = cat_perturb_fn(rng_key, num_categories, n_samples)
        return next_key, sampled
    
    # cont_masks, immut_masks, num_categories = feats_info
    key_1, key_2 = jrand.split(rng_key)
    perturbed_cont = cont_masks * hyper_sphere_coordindates(
        key_1, x, n_samples, high, low, p_norm
    )
    cat_masks = jnp.where(cont_masks, 0, 1)
    perturbed_cat = cat_masks * jnp.concatenate([
        perturb_cat_feat(key_2, num_cat)[1] for num_cat in num_categories
    ], axis=1)

    perturbed = jnp.where(
        immut_masks,
        jnp.repeat(x, n_samples, axis=0),
        perturbed_cont + perturbed_cat
    )
    
    return perturbed

# %% ../../nbs/methods/05_sphere.ipynb 9
def features_to_infos_and_perturb_fn(
    features: FeaturesList
) -> Tuple[List[Array,Array,Array,Array,Array], Callable]:
    cont_masks = []
    immut_masks = []
    n_categories = []
    cat_transformation_name = None
    for (start, end), feat in zip(features.feature_indices, features):
        if feat.is_categorical:
            cont_mask = jnp.zeros(feat.transformation.num_categories)
            immut_mask = jnp.ones_like(cont_mask) * np.array([feat.is_immutable], dtype=np.int32)
            n_categorie = feat.transformation.num_categories
            cat_transformation_name = feat.transformation.name
        else:
            cont_mask = jnp.ones(1)
            immut_mask = cont_mask * np.array([feat.is_immutable], dtype=np.int32)
            n_categorie = 1
        
        cont_masks, immut_masks, n_categories = map(lambda x, y: x + [y], 
            [cont_masks, immut_masks, n_categories],
            [cont_mask, immut_mask, n_categorie]
        )
    
    cont_masks, immut_masks = map(lambda x: jnp.concatenate(x, axis=0), [cont_masks, immut_masks])
    return (cont_masks, immut_masks, tuple(n_categories)), cat_perturb_fn(cat_transformation_name)

def cat_perturb_fn(transformation):
    def ohe_perturb_fn(rng_key, num_categories, n_samples):
        sampled_cat = sample_categorical(rng_key, num_categories, n_samples)
        return jax.nn.one_hot(
            sampled_cat.reshape(-1), num_classes=num_categories
        )
    
    def ordinal_perturb_fn(rng_key, num_categories, n_samples):
        return sample_categorical(
            rng_key, num_categories, n_samples
        )
    
    if transformation == 'ohe':         return ohe_perturb_fn
    elif transformation == 'ordinal':   return ordinal_perturb_fn
    else:                               return sample_categorical


# %% ../../nbs/methods/05_sphere.ipynb 11
@ft.partial(jit, static_argnums=(3, 4, 5, 6, 7, 8, 9))
def _growing_spheres(
    rng_key: jrand.PRNGKey, # Random number generator key
    y_target: Array, # Target label
    x: Array, # Input instance. Shape: (n_features)
    pred_fn: Callable, # Prediction function
    n_steps: int, # Number of steps
    n_samples: int,  # Number of samples to sample
    step_size: float, # Step size
    p_norm: int, # Norm
    perturb_fn: Callable, # Perturbation function
    apply_constraints_fn: Callable, # Apply immutable constraints
    dtype: jnp.dtype = jnp.float32, # Data type
): 
    @jit
    def dist_fn(x, cf):
        if p_norm == 1:
            return jnp.abs(cf - x).sum(axis=1)
        elif p_norm == 2:
            return jnp.linalg.norm(cf - x, ord=2, axis=1)
        else:
            raise ValueError("Only p_norm = 1 or 2 is supported")
    
    @loop_tqdm(n_steps)
    def step(i, state):
        candidate_cf, count, rng_key = state
        rng_key, subkey = jrand.split(rng_key)
        low, high = step_size * count, step_size * (count + 1)
        # Sample around x
        candidates = perturb_fn(rng_key, x, n_samples, high=high, low=low, p_norm=p_norm)
        
        # Apply immutable constraints
        candidates = apply_constraints_fn(x, candidates, hard=True)
        # assert candidates.shape[1] == x.shape[1], f"candidates.shape = {candidates.shape}, x.shape = {x.shape}"

        # Calculate distance
        dist = dist_fn(x, candidates)

        # Calculate counterfactual labels
        candidate_preds = pred_fn(candidates).argmax(axis=1, keepdims=True)
        indices = candidate_preds == y_target

        # Select valid candidates and their distances
        candidates, dist = jax.tree_util.tree_map(
            lambda x: jnp.where(indices, x, jnp.ones_like(x) * jnp.inf), 
            (candidates, dist)
        )

        closest_idx = dist.argmin()
        candidate_cf_update = candidates[closest_idx].reshape(1, -1)

        candidate_cf = jnp.where(
            dist[closest_idx].mean() < dist_fn(x, candidate_cf).mean(),
            candidate_cf_update, 
            candidate_cf
        )
        return candidate_cf, count + 1, subkey
    
    y_target = y_target.reshape(1, -1).argmax(axis=1)
    candidate_cf = jnp.ones_like(x) * jnp.inf
    count = 0
    state = (candidate_cf, count, rng_key)
    candidate_cf, _, _ = lax.fori_loop(0, n_steps, step, state)
    # if `inf` is found, return the original input
    candidate_cf = jnp.where(jnp.isinf(candidate_cf), x, candidate_cf)
    return candidate_cf

# %% ../../nbs/methods/05_sphere.ipynb 12
class GSConfig(BaseConfig):
    n_steps: int = 100
    n_samples: int = 100
    step_size: float = 0.05
    p_norm: int = 2


# %% ../../nbs/methods/05_sphere.ipynb 13
class GrowingSphere(CFModule):
    def __init__(self, config: dict | GSConfig = None, *, name: str = None, perturb_fn = None):
        if config is None:
             config = GSConfig()
        config = validate_configs(config, GSConfig)
        name = "GrowingSphere" if name is None else name
        self.perturb_fn = perturb_fn
        super().__init__(config, name=name)

    def has_data_module(self):
        return hasattr(self, 'data_module') and self.data_module is not None
    
    def save(self, path: str, *, save_data_module: bool = True):
        self.config.save(Path(path) / 'config.json')
        if self.has_data_module() and save_data_module:
            self.data_module.save(Path(path) / 'data_module')
    
    @classmethod
    def load_from_path(cls, path: str):
        config = GSConfig.load_from_json(Path(path) / 'config.json')
        gs = cls(config=config)
        if (Path(path) / 'data_module').exists():
            dm = DataModule.load_from_path(Path(path) / 'data_module')
            gs.set_data_module(dm)
        return gs

    def before_generate_cf(self, *args, **kwargs):
        if self.perturb_fn is None:
            if self.has_data_module():
                feats_info, perturb_fn = features_to_infos_and_perturb_fn(self.data_module.features)
                cont_masks, immut_masks, num_categories = feats_info
                self.perturb_fn = ft.partial(
                    perturb_function_with_features, 
                    cont_masks=cont_masks,
                    immut_masks=immut_masks,
                    num_categories=num_categories,
                    cat_perturb_fn=perturb_fn
                )
                # self.apply_constraints = default_apply_constraints_fn
            else:
                self.perturb_fn = default_perturb_function
        
    @auto_reshaping('x')
    def generate_cf(
        self,
        x: Array,  # `x` shape: (k,), where `k` is the number of features
        pred_fn: Callable[[Array], Array],
        y_target: Array = None,
        rng_key: jnp.ndarray = None,
        **kwargs,
    ) -> jnp.DeviceArray:
        # TODO: Currently assumes binary classification.
        if y_target is None:
            y_target = 1 - pred_fn(x)
        else:
            y_target = y_target.reshape(1, -1)
        if rng_key is None:
            raise ValueError("`rng_key` must be provided, but got `None`.")
        
        return _growing_spheres(
            rng_key=rng_key,
            x=x,
            y_target=y_target,
            pred_fn=pred_fn,
            n_steps=self.config.n_steps,
            n_samples=self.config.n_samples,
            step_size=self.config.step_size,
            p_norm=self.config.p_norm,
            perturb_fn=self.perturb_fn,
            apply_constraints_fn=self.apply_constraints,
        )



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/legacy/04a_logger.ipynb.

# %% ../../nbs/legacy/04a_logger.ipynb 3
from __future__ import annotations
from .import_essentials import *
from collections import defaultdict

# %% auto 0
__all__ = ['Logger']

# %% ../../nbs/legacy/04a_logger.ipynb 4
# class TensorboardLogger:
class Logger:
    _last_logs: Dict = dict()

    def __init__(self, log_dir: Union[str, Path], name: str, on_step: bool = False):
        self._log_dir = log_dir
        self._on_step = on_step
        # self.experiment = test_tube.Experiment(name=name, save_dir=log_dir)
        self.experiment = None
        self._epoch_logs = defaultdict(list)

    @property
    def log_dir(self):
        return self._log_dir
        # name = self.experiment.name
        # version = self.experiment.version
        # return self.experiment.get_data_path(name, version)

    def save_hyperparams(self, hparams: dict | BaseParser):
        pass
        # if issubclass(type(hparams), BaseParser): 
        #     hparams = hparams.dict()
        # if isinstance(hparams, dict):
        #     self.experiment.tag(hparams)
        # else:
        #     raise ValueError(f"hparams should be either `dict`, or a sublcass of `BaseParser`",
        #         f"but got {type(hparams)}.")            
        

    def log(self, name: str, value: Any):
        self.log_dict({name: value})

    def log_dict(self, dictionary: Dict[str, float]):
        for k, v in dictionary.items():
            self._epoch_logs[k].append(v)

        # log to test_tube if on_step is True
        # if self._on_step:
        #     self.experiment.log(dictionary)
        self._last_logs = dictionary

    def get_last_logs(self):
        return self._last_logs

    def on_epoch_started(self):
        self._epoch_logs = defaultdict(list)

    def on_epoch_finished(self):
        epoch_logs = {f"{k}_epoch": np.mean(v) for k, v in self._epoch_logs.items()}
        # self.experiment.log(epoch_logs)
        return epoch_logs

    def close(self):
        pass
        # self.experiment.save()
        # self.experiment.close()



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03_explain.strategy.ipynb.

# %% ../nbs/03_explain.strategy.ipynb 2
from __future__ import annotations
from .import_essentials import *
import einops

# %% auto 0
__all__ = ['BaseStrategy', 'IterativeStrategy', 'VmapStrategy', 'PmapStrategy', 'BatchedVmapStrategy', 'BatchedPmapStrategy',
           'StrategyFactory']

# %% ../nbs/03_explain.strategy.ipynb 3
class BaseStrategy:
    """Base class for mapping strategy."""
    
    def __call__(
        self, 
        fn: Callable, # Function to generate cf for a single input
        xs: Array, # Input instances to be explained
        pred_fn: Callable[[Array], Array],
        y_targets: Array,
        rng_keys: Iterable[jrand.PRNGKey],
        **kwargs
    ) -> Array: # Generated counterfactual explanations
        raise NotImplementedError
    
    __ALL__ = ["__call__"]


# %% ../nbs/03_explain.strategy.ipynb 4
class IterativeStrategy(BaseStrategy):
    """Iterativly generate counterfactuals."""

    def __call__(
        self, 
        fn: Callable, # Function to generate cf for a single input
        xs: Array, # Input instances to be explained
        pred_fn: Callable[[Array], Array],
        y_targets: Array,
        rng_keys: Iterable[jrand.PRNGKey],
        **kwargs
    ) -> Array: # Generated counterfactual explanations
        
        assert xs.ndim == 2
        cfs = jnp.stack([fn(xs[i], pred_fn=pred_fn, y_target=y_targets[i], rng_key=rng_keys[i], **kwargs) 
            for i in range(xs.shape[0])])
        return cfs


# %% ../nbs/03_explain.strategy.ipynb 5
class VmapStrategy(BaseStrategy):
    """Generate counterfactuals via `jax.vmap`."""

    def __call__(
        self, 
        fn: Callable, # Function to generate cf for a single input
        xs: Array, # Input instances to be explained
        pred_fn: Callable[[Array], Array],
        y_targets: Array,
        rng_keys: Iterable[jrand.PRNGKey],
        **kwargs
    ) -> Array: # Generated counterfactual explanations
        
        def partial_fn(x, y_target, rng_key):
            return fn(x, pred_fn=pred_fn, y_target=y_target, rng_key=rng_key, **kwargs)
        
        assert xs.ndim == 2
        cfs = jax.vmap(partial_fn)(xs, y_targets, rng_keys)
        return cfs


# %% ../nbs/03_explain.strategy.ipynb 6
def _pad_divisible_X(
    xs: Array,
    n_devices: int
):
    """Pad `X` to be divisible by `n_devices`."""
    if xs.shape[0] % n_devices != 0:
        pad_size = n_devices - xs.shape[0] % n_devices
        xs_pad = einops.repeat(
            xs[-1:], "n ... -> (pad n) ...", pad=pad_size
        )
        xs = jnp.concatenate([xs, xs_pad])
    X_padded = xs.reshape(n_devices, -1, *xs.shape[1:])
    return X_padded


# %% ../nbs/03_explain.strategy.ipynb 8
class PmapStrategy(BaseStrategy):
    def __init__(
        self, 
        n_devices: int = None, # Number of devices. If None, use all available devices
        strategy: str = 'auto', # Strategy to generate counterfactuals
        **kwargs
    ):
        self.strategy = strategy
        self.n_devices = n_devices or jax.device_count()

    def __call__(
        self, 
        fn: Callable, # Function to generate cf for a single input
        xs: Array, # Input instances to be explained
        pred_fn: Callable[[Array], Array],
        y_targets: Array,
        rng_keys: Iterable[jrand.PRNGKey],
        **kwargs
    ) -> Array: # Generated counterfactual explanations
        
        def partial_fn(x, y_target, rng_key, **kwargs):
            return fn(x, pred_fn=pred_fn, y_target=y_target, rng_key=rng_key, **kwargs)

        assert xs.ndim == 2
        X_padded = _pad_divisible_X(xs, self.n_devices)
        y_targets = _pad_divisible_X(y_targets, self.n_devices)
        rng_keys = _pad_divisible_X(rng_keys, self.n_devices)
        cfs = jax.pmap(jax.vmap(partial_fn))(X_padded, y_targets, rng_keys)
        cfs = cfs.reshape(-1, *cfs.shape[2:])
        cfs = cfs[:xs.shape[0]]
        return cfs


# %% ../nbs/03_explain.strategy.ipynb 9
def _pad_xs(
    xs: Array, pad_size: int, batch_size: int
):
    """Pad `X` to be divisible by `n_devices`."""
    xs_pad = einops.repeat(
        xs[-1:], "n ... -> (pad n) ...", pad=pad_size
    )
    xs = jnp.concatenate([xs, xs_pad])
    xs = einops.rearrange(xs, "(b n) ... -> b n ...", b=batch_size)
    return xs

def _batched_generation(
    gs_fn: Callable, # Generation strategy function
    cf_fn: Callable, # Function to generate cf for a single input
    xs: Array, # Input instances to be explained
    pred_fn: Callable[[Array], Array],
    y_targets: Array,
    rng_keys: Iterable[jrand.PRNGKey],
    batch_size: int,
    **kwargs
) -> Array: # Generated counterfactual explanations
    """Batched  of counterfactuals."""

    def gs_fn_partial(state):
        x, y_target, rng_key = state
        return gs_fn(cf_fn, x, pred_fn, y_target, rng_key, **kwargs)
    
    assert xs.ndim == 2, f"X must be a 2D array, got {xs.ndim}D array"
    x_shape = xs.shape
    batch_size = min(batch_size, x_shape[0])
    # pad X to be divisible by batch_size
    pad_size = batch_size - (xs.shape[0] % batch_size)
    xs = _pad_xs(xs, pad_size, batch_size)
    y_targets = _pad_xs(y_targets, pad_size, batch_size)
    rng_keys = _pad_xs(rng_keys, pad_size, batch_size)
    # generate cfs via lax.map
    cfs = lax.map(gs_fn_partial, (xs, y_targets, rng_keys))
    # cfs = cfs.reshape(-1, *x_shape[1:])[:x_shape[0]]
    cfs = einops.rearrange(cfs, "n b ... k -> (n b) ... k")
    cfs = cfs[:x_shape[0]]
    return cfs

# %% ../nbs/03_explain.strategy.ipynb 10
class BatchedVmapStrategy(BaseStrategy):
    """Auto-batching for generate counterfactuals via `jax.vmap`."""
    def __init__(self, batch_size: int):
        self.batch_size = batch_size

    def __call__(
        self, 
        fn: Callable, # Function to generate cf for a single input
        xs: Array, # Input instances to be explained
        pred_fn: Callable[[Array], Array],
        y_targets: Array,
        rng_keys: Iterable[jrand.PRNGKey],
        **kwargs
    ) -> Array: # Generated counterfactual explanations
        vmap_g = VmapStrategy()    
        cfs = _batched_generation(
            vmap_g, fn, xs, pred_fn, y_targets, rng_keys, self.batch_size, **kwargs
        )
        return cfs


# %% ../nbs/03_explain.strategy.ipynb 11
class BatchedPmapStrategy(BaseStrategy):
    """Auto-batching for generate counterfactuals via `jax.vmap`."""
    def __init__(self, batch_size: int, n_devices: int = None):
        self.batch_size = batch_size
        self.n_devices = n_devices

    def __call__(
        self, 
        fn: Callable, # Function to generate cf for a single input
        xs: Array, # Input instances to be explained
        pred_fn: Callable[[Array], Array],
        y_targets: Array,
        rng_keys: Iterable[jrand.PRNGKey],
        **kwargs
    ) -> Array: # Generated counterfactual explanations
        pmap_g = PmapStrategy(self.n_devices)
        cfs = _batched_generation(
            pmap_g, fn, xs, pred_fn, y_targets, rng_keys, self.batch_size, **kwargs
        )
        return cfs


# %% ../nbs/03_explain.strategy.ipynb 22
class StrategyFactory(object):
    """Factory class for Parallelism Strategy."""

    __strategy_map = {
        'iter': IterativeStrategy(),
        'vmap': VmapStrategy(),
        'pmap': PmapStrategy(),
    }

    def __init__(self) -> None:
        raise ValueError("This class should not be instantiated.")
        
    @staticmethod
    def get_default_strategy() -> BaseStrategy:
        """Get default strategy."""
        return VmapStrategy()

    @classmethod
    def get_strategy(cls, strategy: str | BaseStrategy) -> BaseStrategy:
        """Get strategy."""
        if isinstance(strategy, BaseStrategy):
            return strategy
        elif isinstance(strategy, str) and strategy in cls.__strategy_map:
            return cls.__strategy_map[strategy]
        else:
            raise ValueError(f"Invalid strategy: {strategy}")
        
    __ALL__ = ["get_default_strategy", "get_strategy"]
